import torch
from torch import nn
# import torch.nn.functional as F
# from segmentation_models_pytorch.base.modules import Flatten, Activation

from torchvision.models.segmentation._utils import _SimpleSegmentationModel
from frame_field_learning import tta_utils


class Flatten(torch.nn.Module):
    def forward(self, x):
        return x.view(x.shape[0], -1)


class Activation(nn.Module):

    def __init__(self, name, **params):

        super().__init__()

        if name is None or name == 'identity':
            self.activation = nn.Identity(**params)
        elif name == 'sigmoid':
            self.activation = nn.Sigmoid()
        elif name == 'softmax2d':
            self.activation = nn.Softmax(dim=1, **params)
        elif name == 'softmax':
            self.activation = nn.Softmax(**params)
        elif name == 'logsoftmax':
            self.activation = nn.LogSoftmax(**params)
        elif callable(name):
            self.activation = name(**params)
        else:
            raise ValueError('Activation should be callable/sigmoid/softmax/logsoftmax/None; got {}'.format(name))

    def forward(self, x):
        return self.activation(x)


def get_out_channels(module):
    if hasattr(module, "out_channels"):
        return module.out_channels
    children = list(module.children())
    i = 1
    out_channels = None
    while out_channels is None and i <= len(children):
        last_child = children[-i]
        out_channels = get_out_channels(last_child)
        i += 1
    # If we get out of the loop but out_channels is None,
    # then the prev child of the parent module will be checked, etc.
    return out_channels


# Need to change this later
def get_encoder(module, x):
    x1 = module.conv1(x)
    x2 = module.conv2(x1)
    x3 = module.conv3(x2)
    x4 = module.conv4(x3)
    x5 = module.conv5(x4)

    return [x, x1, x2, x3, x4, x5]


def get_encoder_channels(module):
    c1 = get_out_channels(module.conv1)
    c2 = get_out_channels(module.conv2)
    c3 = get_out_channels(module.conv3)
    c4 = get_out_channels(module.conv4)
    c5 = get_out_channels(module.conv5)

    return [3, c1, c2, c3, c4, c5]


class Multi_FrameFieldModel(torch.nn.Module):
    def __init__(self, config: dict, backbone, train_transform=None, eval_transform=None):
        """

        :param config:
        :param backbone: A _SimpleSegmentationModel network,
                         its output features will be used to compute seg and framefield.
        :param train_transform: transform applied to the inputs when self.training is True
        :param eval_transform: transform applied to the inputs when self.training is False
        """
        super(Multi_FrameFieldModel, self).__init__()
        assert config["compute_seg"] or config["compute_crossfield"], \
            "Model has to compute at least one of those:\n" \
            "\t- segmentation\n" \
            "\t- cross-field"
        assert isinstance(backbone, _SimpleSegmentationModel), \
            "backbone should be an instance of _SimpleSegmentationModel"
        self.config = config
        self.backbone = backbone
        self.train_transform = train_transform
        self.eval_transform = eval_transform

        bn_momentum = 0.1

        backbone_out_features = get_out_channels(self.backbone)

        # For geo-centric pose
        encoder_out_channels = get_encoder_channels(self.backbone.backbone)

        self.xydir_head = EncoderRegressionHead(
            in_channels=encoder_out_channels[-1],
            out_channels=2,
        )

        self.height_head = RegressionHead(
            in_channels=backbone_out_features,
            out_channels=1,
            kernel_size=3,
        )

        self.mag_head = RegressionHead(
            in_channels=backbone_out_features,
            out_channels=1,
            kernel_size=3,
        )

        self.scale_head = ScaleHead()

        # self.initialize_geopose()

        # --- Add other modules if activated in config:
        seg_channels = 0
        if self.config["compute_seg"]:
            seg_channels = self.config["seg_params"]["compute_vertex"]\
                           + self.config["seg_params"]["compute_edge"]\
                           + self.config["seg_params"]["compute_interior"]
            self.seg_module = torch.nn.Sequential(
                torch.nn.Conv2d(backbone_out_features, backbone_out_features, 3, padding=1),
                torch.nn.BatchNorm2d(backbone_out_features, momentum=bn_momentum),
                torch.nn.ELU(),
                torch.nn.Conv2d(backbone_out_features, seg_channels, 1),
                torch.nn.Sigmoid(),)

        if self.config["compute_crossfield"]:
            crossfield_channels = 4
            self.crossfield_module = torch.nn.Sequential(
                torch.nn.Conv2d(backbone_out_features + seg_channels, backbone_out_features, 3, padding=1),
                torch.nn.BatchNorm2d(backbone_out_features, momentum=bn_momentum),
                torch.nn.ELU(),
                torch.nn.Conv2d(backbone_out_features, crossfield_channels, 1),
                torch.nn.Tanh(),
            )

        # Add the entry in the config file later
        if True:
            depth_channels = 1

            self.depth_module = torch.nn.Sequential(

                    torch.nn.Conv2d(backbone_out_features, backbone_out_features * 2, 3, padding=1, bias=False),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, backbone_out_features * 2, 3, padding=1, bias=False),
                    torch.nn.BatchNorm2d(backbone_out_features * 2, momentum=bn_momentum),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, backbone_out_features * 2, 3, padding=1, bias=False),
                    torch.nn.BatchNorm2d(backbone_out_features * 2, momentum=bn_momentum),
                    torch.nn.ReLU(),

                    # This is original
                    torch.nn.Conv2d(backbone_out_features * 2, depth_channels, 1, bias=True),
                    # torch.nn.ReLU(),

                    # torch.nn.Conv2d(backbone_out_features*2, backbone_out_features*2, 3, padding=1, bias=False),

                    # torch.nn.BatchNorm2d(backbone_out_features*2),
                    # torch.nn.ReLU(),

                    # torch.nn.Conv2d(backbone_out_features*2, depth_channels, 1, bias=True),
                    # torch.nn.ReLU(),
            )

            # shadow_channels = 1

            self.shadow_module = torch.nn.Sequential(
                    torch.nn.Conv2d(backbone_out_features, backbone_out_features * 2, 3, padding=1, bias=True),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, backbone_out_features * 2, 3, padding=1, bias=False),
                    torch.nn.BatchNorm2d(backbone_out_features * 2, momentum=bn_momentum),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, backbone_out_features * 2, 3, padding=1, bias=False),
                    torch.nn.BatchNorm2d(backbone_out_features * 2, momentum=bn_momentum),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, depth_channels, 1, bias=True),
                    torch.nn.Sigmoid(),
                    # torch.nn.Tanh(),
            )

            # facade_channels = 1

            self.facade_module = torch.nn.Sequential(
                    torch.nn.Conv2d(backbone_out_features, backbone_out_features * 2, 3, padding=1, bias=True),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, backbone_out_features * 2, 3, padding=1, bias=False),
                    torch.nn.BatchNorm2d(backbone_out_features * 2, momentum=bn_momentum),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, backbone_out_features * 2, 3, padding=1, bias=False),
                    torch.nn.BatchNorm2d(backbone_out_features * 2, momentum=bn_momentum),
                    torch.nn.ReLU(),

                    torch.nn.Conv2d(backbone_out_features * 2, depth_channels, 1, bias=True),
                    torch.nn.Sigmoid(),
                    # torch.nn.Tanh(),
            )

    '''
    def initialize_geopose(self):
        init.initialize_decoder(self.decoder)
        init.initialize_head(self.xydir_head)
        init.initialize_head(self.height_head)
        init.initialize_head(self.mag_head)
        init.initialize_head(self.scale_head)
    '''

    def inference(self, image):
        outputs = {}

        encoder_features = get_encoder(self.backbone.backbone, image)

        # --- Extract features for every pixel of the image with a U-Net --- #
        backbone_features0 = self.backbone(image)["out"]

        if self.config["compute_seg"]:
            # --- Output a segmentation of the image --- #
            seg = self.seg_module(backbone_features0)
            seg_to_cat = seg.clone().detach()
            backbone_features = torch.cat([backbone_features0, seg_to_cat], dim=1)  # Add seg to image features
            outputs["seg"] = seg

        if self.config["compute_crossfield"]:
            # --- Output a cross-field of the image --- #
            crossfield = 2 * self.crossfield_module(backbone_features)  # Outputs c_0, c_2 values in [-2, 2]
            outputs["crossfield"] = crossfield

        # Add the entry in the config file later
        if True:
            # --- Output a depth map of the image --- #
            depth = self.depth_module(backbone_features0)  # Outputs values in [0, ???]
            outputs["depth"] = depth

            # --- Output a shadow response map of the image --- #
            shadow = self.shadow_module(backbone_features0)  # Outputs values in [0, 1]
            outputs["shadow"] = shadow

            # --- Output a facade response map of the image --- #
            facade = self.facade_module(backbone_features0)  # Outputs values in [0, 1]
            outputs["facade"] = facade

        if True:

            # For Geo-Centric Pose
            xydir = self.xydir_head(encoder_features[-1])
            height = self.height_head(backbone_features0)
            # height = depth
            mag = self.mag_head(backbone_features0)
            scale = self.scale_head(mag, height)

            if scale.ndim == 0:
                scale = torch.unsqueeze(scale, axis=0)

            outputs['xydir'] = xydir
            outputs['height'] = height
            outputs['mag'] = mag
            outputs['scale'] = scale

        return outputs

    # @profile
    def forward(self, xb, tta=False):

        # pdb.set_trace()

        # print("\n### --- PolyRefine.forward(xb) --- ####")
        if self.training:
            if self.train_transform is not None:
                xb = self.train_transform(xb)
        else:
            if self.eval_transform is not None:
                xb = self.eval_transform(xb)

        if not tta:
            final_outputs = self.inference(xb["image"])
        else:
            final_outputs = tta_utils.tta_inference(self, xb, self.config["eval_params"]["seg_threshold"])

        return final_outputs, xb


class RegressionHead(torch.nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3):
        conv2d = torch.nn.Conv2d(in_channels, out_channels, kernel_size, padding=kernel_size // 2)
        identity = torch.nn.Identity()
        activation = Activation(None)
        super().__init__(conv2d, identity, activation)


class EncoderRegressionHead(torch.nn.Sequential):
    def __init__(self, in_channels, out_channels, kernel_size=3):
        pool = torch.nn.AdaptiveAvgPool2d(1)
        flatten = Flatten()
        dropout = torch.nn.Dropout(p=0.5, inplace=True)
        linear = torch.nn.Linear(in_channels, 2, bias=True)
        super().__init__(pool, flatten, dropout, linear)


class ScaleHead(torch.nn.Module):
    def __init__(
        self,
    ):
        super().__init__()
        self.flatten = torch.flatten
        self.dot = torch.dot

    def forward(self, mag, height):
        curr_mag = self.flatten(mag, start_dim=1)
        curr_height = self.flatten(height, start_dim=1)
        batch_size = curr_mag.shape[0]
        length = curr_mag.shape[1]
        denom = (
            torch.squeeze(
                torch.bmm(
                    curr_height.view(batch_size, 1, length),
                    curr_height.view(batch_size, length, 1),
                )
            )
            + 0.01
        )
        pinv = curr_height / denom.view(batch_size, 1)
        scale = torch.squeeze(
            torch.bmm(
                pinv.view(batch_size, 1, length), curr_mag.view(batch_size, length, 1)
            )
        )
        return scale



"""
Print stats about a torch model

Exapmle Usage:
    DVC_DPATH=$(geowatch_dvc)
    PACKAGE_FPATH=$DVC_DPATH/models/fusion/SC-20201117/BOTH_smt_it_stm_p8_L1_DIL_v55/BOTH_smt_it_stm_p8_L1_DIL_v55_epoch=5-step=53819.pt
    python -m geowatch.cli.torch_model_stats $PACKAGE_FPATH
"""
import scriptconfig as scfg
import ubelt as ub


class TorchModelStatsConfig(scfg.DataConfig):
    """
    Print stats about a torch model.

    Currently some things are hard-coded for fusion models
    """
    src = scfg.PathList(help='path to one or more torch models', position=1)
    stem_stats = scfg.Value(True, isflag=True, help='if True, print more verbose model mean/std')
    hparams = scfg.Value(True, isflag=True, help='if True, print fit hyperparameters')


def main(cmdline=False, **kwargs):
    """
    Ignore:
        import geowatch
        from geowatch.cli.torch_model_stats import *  # NOQA
        dvc_dpath = geowatch.find_dvc_dpath()
        package_fpath1 = dvc_dpath / 'models/fusion/SC-20201117/BOTH_smt_it_stm_p8_L1_DIL_v55/BOTH_smt_it_stm_p8_L1_DIL_v55_epoch=5-step=53819.pt'
        package_fpath2 = dvc_dpath / 'models/fusion/SC-20201117/BAS_smt_it_stm_p8_TUNE_L1_RAW_v58/BAS_smt_it_stm_p8_TUNE_L1_RAW_v58_epoch=3-step=81135.pt'
        package_fpath = dvc_dpath / 'models/fusion/SC-20201117/BAS_smt_it_stm_p8_L1_raw_v53/BAS_smt_it_stm_p8_L1_raw_v53_epoch=3-step=85011.pt'
        kwargs = {
            'src': [package_fpath2, package_fpath1]
        }
        main(cmdline=False, **kwargs)

    """
    config = TorchModelStatsConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    from rich.markup import escape
    rich.print('config = {}'.format(escape(ub.urepr(config, nl=1))))
    package_paths = config['src']

    import geowatch
    import warnings
    if not ub.iterable(package_paths):
        package_paths = [package_paths]

    try:
        dvc_dpath = geowatch.find_dvc_dpath()
    except Exception:
        dvc_dpath = None

    package_rows = []
    for package_fpath in package_paths:
        print('--------')
        print(f'package_fpath={package_fpath}')

        stem_stats = config['stem_stats']
        try:
            row = torch_model_stats(package_fpath, stem_stats=stem_stats, dvc_dpath=dvc_dpath)
        except RuntimeError as pkg_ex:
            print('Reading as a package failed, attempting to read as a checkpoint')
            checkpoint_fpath = package_fpath
            try:
                checkpoint_stats = torch_checkpoint_stats(checkpoint_fpath)
            except Exception as ckpt_ex:
                warnings.warn(f'Unable to read input as a checkpoint due to: {ckpt_ex}.')
                raise pkg_ex
            else:
                warnings.warn(f'Unable to read input as a package due to: {pkg_ex}. Interpreting as a checkpoint instead')

            rich.print('checkpoint_stats = {}'.format(ub.urepr(checkpoint_stats, nl=2, sort=0, precision=2)))
        else:
            model_stats = row.get('model_stats', None)
            fit_config = row.pop('fit_config', None)
            config_cli_yaml = row.pop('config_cli_yaml', None)
            if config.hparams:
                rich.print('fit_config = {}'.format(ub.urepr(fit_config, nl=1)))
                rich.print('config_cli_yaml = {}'.format(ub.urepr(config_cli_yaml, nl=2)))
            rich.print('model_stats = {}'.format(ub.urepr(model_stats, nl=2, sort=0, precision=2)))
            package_rows.append(row)

    print('package_rows = {}'.format(ub.urepr(package_rows, nl=2, sort=0)))


def torch_checkpoint_stats(checkpoint_fpath):
    """
    A fallback set of statistics we can make for checkpoints only.

    Summarizes a PyTorch checkpoint by extracting useful metadata about the model's state_dict
    and optimizer states.

    Args:
        checkpoint_path (str | PathLike): Path to the checkpoint file.

    Returns:
        dict: Summary statistics of the checkpoint.
    """
    import torch
    data = torch.load(checkpoint_fpath)
    top_level_keys = list(data.keys())
    print(f'top_level_keys = {ub.urepr(top_level_keys, nl=1)}')

    # TODO: Given a checkpoint path, we may be able to read more information
    # about it if we can find the corresponding haprams.yaml or
    # train_config.yaml

    nested_keys = {
        'state_dict',
        'optimizer_states',
        'loops'
    }
    nested = ub.udict.intersection(data, nested_keys)
    non_nested = ub.udict.difference(data, nested_keys)
    stats = {
        'path': checkpoint_fpath,
        **non_nested
    }

    if 'optimizer_states' in nested:
        optimizer_states_stats = {}
        optimizer_states = nested['optimizer_states']
        optimizer_states_stats['num_optimizer_states'] = len(optimizer_states)
        opt_stat_list = []
        for opt_state in optimizer_states:
            opt_state_stats = {}
            # TODO: Can we do more here?
            opt_state_stats['num_param_groups'] = len(opt_state['param_groups'])
            opt_stat_list.append(opt_state_stats)
        optimizer_states_stats['total_num_param_groups'] = sum(r['num_param_groups'] for r in opt_stat_list)
    else:
        optimizer_states_stats = None

    if 'state_dict' in nested:
        state_dict_stats = {}
        state_dict = nested['state_dict']
        len(state_dict)
        # state_shape = ub.udict.map_values(state_dict, lambda x: x.shape)
        state_numel = ub.udict.map_values(state_dict, lambda x: x.numel())
        total_params = sum(state_numel.values())
        # TODO: Can we use the optimizer to determine the trainable params?
        state_dict_stats['num_tensors'] = len(state_dict)
        state_dict_stats['total_params'] = total_params
    else:
        state_dict_stats = None

    if 'loops' in nested:
        loop_stats = {}
        # TODO: extract appopriate stats
        loops = nested['loops']
    else:
        loops = None
        loop_stats = None
    # Just use the loops
    loop_stats = loops

    stats['loops'] = loop_stats
    stats['state_dict'] = state_dict_stats
    stats['optimizer_states_stats'] = optimizer_states_stats
    return stats


def torch_model_stats(package_fpath, stem_stats=True, dvc_dpath=None):
    import kwcoco
    from geowatch.tasks.fusion import utils
    from geowatch.utils import util_netharn
    from geowatch.monkey import monkey_torchmetrics
    monkey_torchmetrics.fix_torchmetrics_compatability()

    package_fpath = ub.Path(package_fpath)

    if not package_fpath.exists():
        if package_fpath.augment(tail='.dvc').exists():
            raise Exception('model does not exist, but its dvc file does')
        else:
            raise Exception('model does not exist')

    file_stat = package_fpath.stat()

    # TODO: generalize the load-package
    raw_module, package_header = utils.load_model_from_package(package_fpath, with_header=True)

    if hasattr(raw_module, 'module'):
        module = raw_module.module
    else:
        module = raw_module

    # TODO: get the category freq

    model_stats = {}
    num_params = util_netharn.number_of_parameters(module)

    print(ub.urepr(utils.model_json(module, max_depth=3), nl=-1, sort=0))
    # print(ub.urepr(utils.model_json(module, max_depth=2), nl=-1, sort=0))

    # import xdev
    # with xdev.embed_on_exception_context:
    try:
        state = module.state_dict()
    except Exception:
        if hasattr(module, 'head_metrics'):
            module.head_metrics.clear()
            state = module.state_dict()
        else:
            raise
    state_keys = list(state.keys())
    # print('state_keys = {}'.format(ub.urepr(state_keys, nl=1)))

    unique_sensors = set()
    config_cli_yaml = None
    train_dataset = None
    prenorm_stats = None
    fit_config = {}
    if hasattr(module, 'dataset_stats') and module.dataset_stats is not None:

        dataset_stats = module.dataset_stats.copy()

        if 'modality_input_stats' in dataset_stats:
            # This is too much info to print
            dataset_stats.pop('modality_input_stats', None)

        known_input_stats = []
        unknown_input_stats = []
        sensor_modes_with_stats = set()

        input_stats = dataset_stats.get('input_stats', None)
        unique_sensor_modes = dataset_stats.get('unique_sensor_modes', None)
        if input_stats is not None:

            if isinstance(input_stats, list):
                # Convert to dictionary form
                input_stats_dict = {}
                for item in input_stats:
                    sens_chan_key = (item['sensor'], item['channels'])
                    input_stats_dict[sens_chan_key] = ub.udict(item) - {'sensor', 'channels'}
                input_stats = input_stats_dict

            for sens_chan_key, stats in input_stats.items():
                sensor, channel = sens_chan_key
                channel = kwcoco.ChannelSpec.coerce(channel).concise().spec
                sensor_modes_with_stats.add((sensor, channel))
                unique_sensors.add(sensor)
                sensor_stat = {
                    'sensor': sensor,
                    'channel': channel,
                }
                if stem_stats:
                    import numpy as np
                    sensor_stat.update({
                        'mean': np.asarray(stats['mean']).ravel().tolist(),
                        'std': np.asarray(stats['std']).ravel().tolist(),
                    })
                known_input_stats.append(sensor_stat)
        else:
            known_input_stats = None

        if unique_sensor_modes is not None:
            unique_sensor_modes = list(dataset_stats['unique_sensor_modes'])
            for sensor, channel in unique_sensor_modes:
                channel = kwcoco.ChannelSpec.coerce(channel).concise().spec
                key = (sensor, channel)
                if key not in sensor_modes_with_stats:
                    unique_sensors.add(sensor)
                    unknown_input_stats.append(
                        {
                            'sensor': sensor,
                            'channel': channel,
                        }
                    )
        else:
            unknown_input_stats = None

        mb_size = file_stat.st_size / (2.0 ** 20)
        size_str = ub.urepr(mb_size, precision=2) + ' MB'

        # Add in some params about how this model was trained
        if hasattr(raw_module, 'config_cli_yaml'):
            config_cli_yaml = raw_module.config_cli_yaml
        else:
            config_cli_yaml = None

        if hasattr(raw_module, 'fit_config'):
            # Old non-cli modules
            fit_config = raw_module.fit_config
        else:
            # new lightning cli modules
            fit_config = (
                ub.udict(getattr(raw_module, 'datamodule_hparams', {})) |
                ub.udict(raw_module.hparams)
            )

        if 'train_dataset' in fit_config:
            train_dataset = ub.Path(fit_config['train_dataset'])
        else:
            if config_cli_yaml is not None:
                train_dataset = config_cli_yaml.get('data', {}).get('train_dataset', None)
            else:
                train_dataset = None

        if dvc_dpath is not None and train_dataset is not None:
            try:
                if str(train_dataset).startswith(str(dvc_dpath)):
                    train_dataset = train_dataset.relative_to(dvc_dpath)

                if str(package_fpath).startswith(str(dvc_dpath)):
                    package_fpath = package_fpath.relative_to(dvc_dpath)
            except Exception:
                ...

        heads = []
        if fit_config['global_class_weight']:
            heads.append('class')

        if fit_config['global_saliency_weight']:
            heads.append('saliency')

        spacetime_stats = ub.udict(fit_config) & [
            'chip_size',
            'time_steps',
            'time_sampling',
            'time_span',
            'chip_dims',
            'window_space_scale',
            'input_space_scale',
        ]

        # spacetime_stats = {
        #     'chip_size': fit_config['chip_size'],
        #     'time_steps': fit_config['time_steps'],
        #     'time_sampling': fit_config['time_sampling'],
        #     'time_span': fit_config['time_span'],
        # }

        model_stats['size'] = size_str
        model_stats['num_params'] = num_params
        model_stats['num_states'] = len(state_keys)
        model_stats['heads'] = heads
        model_stats['train_dataset'] = None if train_dataset is None else str(train_dataset)
        model_stats['spacetime_stats'] = spacetime_stats
        model_stats['classes'] = list(module.classes)
        model_stats['known_inputs'] = known_input_stats
        model_stats['unknown_inputs'] = unknown_input_stats

        # Normalization done in the dataloader
        prenorm_stats = {
            'normalize_peritem': fit_config.get('normalize_peritem'),
        }

    param_stats = {
        name: {
            "size": param.size(),
            "min": param.min().item(),
            "max": param.max().item(),
            "mean": param.mean().item(),
            "std": param.std().item(),
        }
        for name, param in module.named_parameters()
    }
    param_stats_summary = {
        "min": min([summary["min"] for summary in param_stats.values()]),
        "max": min([summary["max"] for summary in param_stats.values()]),
        "mean": min([summary["mean"] for summary in param_stats.values()]),
    }

    try:
        unique_sensors = sorted(unique_sensors)
    except TypeError:
        ...

    row = {
        'name': package_fpath.stem,
        'task': 'TODO',
        'file_name': str(package_fpath),
        'sensors': unique_sensors,
        'train_dataset': str(train_dataset),
        'fit_config': fit_config,
        'config_cli_yaml': config_cli_yaml,
        'model_stats': model_stats,
        'prenorm_stats': prenorm_stats,
        'param_stats': param_stats_summary,
        'package_header': package_header,
    }

    if hasattr(module, 'input_sensorchan'):
        input_sensorchan = module.input_sensorchan.concise().spec
        row['input_sensorchan'] = input_sensorchan
    elif hasattr(module, 'input_channels'):
        input_channels = module.input_channels.concise().spec
        row['input_channels'] = input_channels

    return row


def fallback(fpath):
    import zipfile
    zfile = zipfile.ZipFile(fpath)
    for internal_path in zfile.namelist():
        if internal_path.endswith('.yaml'):
            data = zfile.read(internal_path)
            print(data.decode('utf8'))


__cli__ = TorchModelStatsConfig

if __name__ == '__main__':
    """
    CommandLine:
        python -m geowatch.cli.torch_model_stats
    """
    main(cmdline=True)



"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/balanced_sampling.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.balanced_sampling import BalancedSampleTree  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.balanced_sampling import BalancedSampleForest  # NOQA



import argparse
import dataclasses
import inspect
import logging
import os
import sys
from contextlib import contextmanager
from contextvars import ContextVar
from typing import (  # type: ignore[attr-defined]
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    _GenericAlias,
)

from ._namespace import Namespace
from ._optionals import (
    get_alias_target,
    get_annotated_base_type,
    import_reconplogger,
    is_alias_type,
    is_annotated,
    reconplogger_support,
)
from ._type_checking import ArgumentParser

__all__ = [
    "LoggerProperty",
    "null_logger",
]

ClassType = TypeVar("ClassType")

if sys.version_info < (3, 8):
    from typing import Callable

    InstantiatorCallable = Callable[..., ClassType]
else:
    from typing import Protocol

    class InstantiatorCallable(Protocol):
        def __call__(self, class_type: Type[ClassType], *args, **kwargs) -> ClassType:
            pass  # pragma: no cover


InstantiatorsDictType = Dict[Tuple[type, bool], InstantiatorCallable]


parent_parser: ContextVar[Optional["ArgumentParser"]] = ContextVar("parent_parser", default=None)
parser_capture: ContextVar[bool] = ContextVar("parser_capture", default=False)
defaults_cache: ContextVar[Optional[Namespace]] = ContextVar("defaults_cache", default=None)
lenient_check: ContextVar[Union[bool, str]] = ContextVar("lenient_check", default=False)
load_value_mode: ContextVar[Optional[str]] = ContextVar("load_value_mode", default=None)
class_instantiators: ContextVar[Optional[InstantiatorsDictType]] = ContextVar("class_instantiators")
nested_links: ContextVar[List[dict]] = ContextVar("nested_links", default=[])
applied_instantiation_links: ContextVar[Optional[set]] = ContextVar("applied_instantiation_links", default=None)


parser_context_vars = dict(
    parent_parser=parent_parser,
    parser_capture=parser_capture,
    defaults_cache=defaults_cache,
    lenient_check=lenient_check,
    load_value_mode=load_value_mode,
    class_instantiators=class_instantiators,
    nested_links=nested_links,
    applied_instantiation_links=applied_instantiation_links,
)


@contextmanager
def parser_context(**kwargs):
    context_var_tokens = []
    for name, value in kwargs.items():
        context_var = parser_context_vars[name]
        token = context_var.set(value)
        context_var_tokens.append((context_var, token))
    try:
        yield
    finally:
        for context_var, token in context_var_tokens:
            context_var.reset(token)


def is_subclass(cls, class_or_tuple) -> bool:
    """Extension of issubclass that supports non-class arguments."""
    try:
        return inspect.isclass(cls) and issubclass(cls, class_or_tuple)
    except TypeError:
        return False


def is_final_class(cls) -> bool:
    """Checks whether a class is final, i.e. decorated with ``typing.final``."""
    return getattr(cls, "__final__", False)


def is_generic_class(cls) -> bool:
    return isinstance(cls, _GenericAlias) and getattr(cls, "__module__", "") != "typing"


def get_generic_origin(cls):
    return cls.__origin__ if is_generic_class(cls) else cls


def get_unaliased_type(cls):
    new_cls = cls
    while True:
        cur_cls = new_cls
        if is_annotated(new_cls):
            new_cls = get_annotated_base_type(new_cls)
        if is_alias_type(new_cls):
            new_cls = get_alias_target(new_cls)
        if new_cls == cur_cls:
            break
    return cur_cls


def is_dataclass_like(cls) -> bool:
    if is_generic_class(cls):
        return is_dataclass_like(cls.__origin__)
    if not inspect.isclass(cls) or cls is object:
        return False
    if is_final_class(cls):
        return True
    classes = [c for c in inspect.getmro(cls) if c not in {object, Generic}]
    all_dataclasses = all(dataclasses.is_dataclass(c) for c in classes)

    if not all_dataclasses:
        from ._optionals import attrs_support, is_pydantic_model

        if is_pydantic_model(cls):
            return True

        if attrs_support:
            import attrs

            if attrs.has(cls):
                return True

    return all_dataclasses


def default_class_instantiator(class_type: Type[ClassType], *args, **kwargs) -> ClassType:
    return class_type(*args, **kwargs)


class ClassInstantiator:
    def __init__(self, instantiators: InstantiatorsDictType) -> None:
        self.instantiators = instantiators

    def __call__(self, class_type: Type[ClassType], *args, **kwargs) -> ClassType:
        for (cls, subclasses), instantiator in self.instantiators.items():
            if class_type is cls or (subclasses and is_subclass(class_type, cls)):
                param_names = set(inspect.signature(instantiator).parameters.keys())
                if "applied_instantiation_links" in param_names:
                    applied_links = applied_instantiation_links.get() or set()
                    kwargs["applied_instantiation_links"] = {
                        action.target[0]: action.applied_value for action in applied_links
                    }
                return instantiator(class_type, *args, **kwargs)
        return default_class_instantiator(class_type, *args, **kwargs)


def get_class_instantiator() -> InstantiatorCallable:
    instantiators = class_instantiators.get()
    if not instantiators:
        return default_class_instantiator
    return ClassInstantiator(instantiators)


# logging

logging_levels = {"CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"}
null_logger = logging.getLogger("jsonargparse_fork_null_logger")
null_logger.addHandler(logging.NullHandler())
null_logger.parent = None


def setup_default_logger(data, level, caller):
    name = caller
    if isinstance(data, str):
        name = data
    elif isinstance(data, dict) and "name" in data:
        name = data["name"]
    logger = logging.getLogger(name)
    logger.parent = None
    if len(logger.handlers) == 0:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
        logger.addHandler(handler)
    level = getattr(logging, level)
    for handler in logger.handlers:
        handler.setLevel(level)
    return logger


def parse_logger(logger: Union[bool, str, dict, logging.Logger], caller):
    if not isinstance(logger, (bool, str, dict, logging.Logger)):
        raise ValueError(f"Expected logger to be an instance of (bool, str, dict, logging.Logger), but got {logger}.")
    if isinstance(logger, dict) and len(set(logger.keys()) - {"name", "level"}) > 0:
        value = {k: v for k, v in logger.items() if k not in {"name", "level"}}
        raise ValueError(f"Unexpected data to configure logger: {value}.")
    if logger is False:
        return null_logger
    level = "WARNING"
    if isinstance(logger, dict) and "level" in logger:
        level = logger["level"]
    if level not in logging_levels:
        raise ValueError(f"Got logger level {level!r} but must be one of {logging_levels}.")
    if (logger is True or (isinstance(logger, dict) and "name" not in logger)) and reconplogger_support:
        kwargs = {"level": "DEBUG", "reload": True} if debug_mode_active() else {}
        logger = import_reconplogger("parse_logger").logger_setup(**kwargs)
    if not isinstance(logger, logging.Logger):
        logger = setup_default_logger(logger, level, caller)
    return logger


class LoggerProperty:
    """Class designed to be inherited by other classes to add a logger property."""

    def __init__(self, *args, logger: Union[bool, str, dict, logging.Logger] = False, **kwargs):
        """Initializer for LoggerProperty class."""
        self.logger = logger  # type: ignore[assignment]
        super().__init__(*args, **kwargs)

    @property
    def logger(self) -> logging.Logger:
        """The logger property for the class.

        :getter: Returns the current logger.
        :setter: Sets the given logging.Logger as logger or sets the default logger
                 if given True/str(logger name)/dict(name, level), or disables logging
                 if given False.

        Raises:
            ValueError: If an invalid logger value is given.
        """
        return self._logger

    @logger.setter
    def logger(self, logger: Union[bool, str, dict, logging.Logger]):
        if logger is None:
            from ._deprecated import deprecation_warning, logger_property_none_message

            deprecation_warning((LoggerProperty.logger, None), logger_property_none_message, stacklevel=2)
            logger = False
        if not logger and debug_mode_active():
            logger = {"level": "DEBUG"}
        self._logger = parse_logger(logger, type(self).__name__)


def debug_mode_active() -> bool:
    return os.getenv("JSONARGPARSE_DEBUG", "").lower() not in {"", "false", "no", "0"}


if debug_mode_active():
    os.environ["LOGGER_LEVEL"] = "DEBUG"  # pragma: no cover


# base classes


class Action(LoggerProperty, argparse.Action):
    """Base for jsonargparse Action classes."""

    def _check_type_(self, value, **kwargs):
        if not hasattr(self, "_check_type_kwargs"):
            self._check_type_kwargs = set(inspect.signature(self._check_type).parameters.keys())
        kwargs = {k: v for k, v in kwargs.items() if k in self._check_type_kwargs}
        return self._check_type(value, **kwargs)



from kwcoco_dataloader.tasks.fusion.datamodules.batch_visualization import BatchVisualizationBuilder  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.batch_visualization import _draw_overlay_item_by_itself  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.batch_visualization import _draw_row_item  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.batch_visualization import _debug_sample_in_context  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.batch_visualization import colorize_weights  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.batch_visualization import _memo_legend  # NOQA



import argparse
import inspect
import locale
import os
import re
import warnings
from collections import defaultdict
from contextlib import contextmanager, suppress
from contextvars import ContextVar
from enum import Enum
from importlib.util import find_spec
from subprocess import PIPE, Popen
from typing import List, Union

from ._actions import ActionConfigFile, _ActionHelpClassPath, remove_actions
from ._parameter_resolvers import get_signature_parameters
from ._typehints import (
    ActionTypeHint,
    callable_origin_types,
    get_all_subclass_paths,
    get_callable_return_type,
    get_typehint_origin,
    is_subclass,
    type_to_str,
)
from ._util import NoneType, Path, import_object, unique

shtab_shell: ContextVar = ContextVar("shtab_shell")
shtab_prog: ContextVar = ContextVar("shtab_prog")
shtab_preambles: ContextVar = ContextVar("shtab_preambles")


def handle_completions(parser):
    if find_spec("argcomplete") and "_ARGCOMPLETE" in os.environ:
        import argcomplete

        from ._common import parser_context

        with parser_context(load_value_mode=parser.parser_mode):
            argcomplete.autocomplete(parser)

    if find_spec("shtab") and not getattr(parser, "parent_parser", None):
        if not any(isinstance(action, ShtabAction) for action in parser._actions):
            parser.add_argument("--print_shtab", action=ShtabAction)


# argcomplete


def get_files_completer():
    from argcomplete.completers import FilesCompleter

    return FilesCompleter()


def argcomplete_namespace(caller, parser, namespace):
    if caller == "argcomplete":
        namespace.__class__ = __import__("jsonargparse_fork").Namespace
        namespace = parser.merge_config(parser.get_defaults(skip_check=True), namespace).as_flat()
    return namespace


def argcomplete_warn_redraw_prompt(prefix, message):
    import argcomplete

    if prefix != "":
        argcomplete.warn(message)
        with suppress(Exception):
            proc = Popen(f"ps -p {os.getppid()} -oppid=".split(), stdout=PIPE, stderr=PIPE)
            stdout, _ = proc.communicate()
            shell_pid = int(stdout.decode().strip())
            os.kill(shell_pid, 28)
    _ = "_" if locale.getlocale()[1] != "UTF-8" else "\xa0"
    return [_ + message.replace(" ", _), ""]


# shtab


class ShtabAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
    ):
        import shtab

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            choices=shtab.SUPPORTED_SHELLS,
            help="Print shtab shell completion script.",
        )

    def __call__(self, parser, namespace, shell, option_string=None):
        import shtab

        warnings.warn("Automatic shtab support is experimental and subject to change.", UserWarning)
        prog = norm_name(parser.prog)
        assert prog
        preambles = []
        if shell == "bash":
            preambles = [bash_compgen_typehint.strip().replace("%s", prog)]
        with prepare_actions_context(shell, prog, preambles):
            shtab_prepare_actions(parser)
        print(shtab.complete(parser, shell, preamble="\n".join(preambles)))
        parser.exit(0)


@contextmanager
def prepare_actions_context(shell, prog, preambles):
    token_shell = shtab_shell.set(shell)
    token_prog = shtab_prog.set(prog)
    token_preambles = shtab_preambles.set(preambles)
    try:
        yield
    finally:
        shtab_shell.reset(token_shell)
        shtab_prog.reset(token_prog)
        shtab_preambles.reset(token_preambles)


def norm_name(name: str) -> str:
    return re.sub(r"\W+", "_", name)


def shtab_prepare_actions(parser) -> None:
    remove_actions(parser, (ShtabAction,))
    if parser._subcommands_action:
        for subparser in parser._subcommands_action._name_parser_map.values():
            shtab_prepare_actions(subparser)
    for action in parser._actions:
        shtab_prepare_action(action, parser)


def shtab_prepare_action(action, parser) -> None:
    import shtab

    if action.choices or hasattr(action, "complete"):
        return

    complete = None
    if isinstance(action, ActionConfigFile):
        complete = shtab.FILE
    elif isinstance(action, ActionTypeHint):
        typehint = action._typehint
        if get_typehint_origin(typehint) == Union:
            subtypes = [s for s in typehint.__args__ if s not in {NoneType, str, dict, list, tuple, bytes}]
            if len(subtypes) == 1:
                typehint = subtypes[0]
        if is_subclass(typehint, Path):
            if "f" in typehint._mode:
                complete = shtab.FILE
            elif "d" in typehint._mode:
                complete = shtab.DIRECTORY
        elif is_subclass(typehint, os.PathLike):
            complete = shtab.FILE
    if complete:
        action.complete = complete
        return

    choices = None
    if isinstance(action, ActionTypeHint):
        skip = getattr(action, "sub_add_kwargs", {}).get("skip", set())
        choices = get_typehint_choices(action._typehint, action.option_strings[0], parser, skip)
        if shtab_shell.get() == "bash":
            message = f"Expected type: {type_to_str(action._typehint)}"
            add_bash_typehint_completion(parser, action, message, choices)
            choices = None
    elif isinstance(action, _ActionHelpClassPath):
        choices = get_help_class_choices(action._baseclass)
    if choices:
        action.choices = choices


bash_compgen_typehint_name = "_jsonargparse_fork_%s_compgen_typehint"
bash_compgen_typehint = """
_jsonargparse_fork_%%s_matched_choices() {
  local TOTAL=$(echo "$1" | wc -w | tr -d " ")
  if [ "$TOTAL" != 0 ]; then
    local MATCH=$(echo "$2" | wc -w | tr -d " ")
    printf "; $MATCH/$TOTAL matched choices"
  fi
}
%(name)s() {
  local MATCH=( $(IFS=" " compgen -W "$1" "$2") )
  if [ ${#MATCH[@]} = 0 ]; then
    if [ "$COMP_TYPE" = 63 ]; then
      MATCHED=$(_jsonargparse_fork_%%s_matched_choices "$1" "${MATCH[*]}")
      printf "%(b)s\\n$3$MATCHED\\n%(n)s" >&2
      kill -WINCH $$
    fi
  else
    IFS=" " compgen -W "$1" "$2"
    if [ "$COMP_TYPE" = 63 ]; then
      MATCHED=$(_jsonargparse_fork_%%s_matched_choices "$1" "${MATCH[*]}")
      printf "%(b)s\\n$3$MATCHED%(n)s" >&2
    fi
  fi
}
""" % {
    "name": bash_compgen_typehint_name,
    "b": "$(tput setaf 5)",
    "n": "$(tput sgr0)",
}


def add_bash_typehint_completion(parser, action, message, choices) -> None:
    fn_typehint = norm_name(bash_compgen_typehint_name % shtab_prog.get())
    fn_name = parser.prog.replace(" [options] ", "_")
    fn_name = norm_name(f"_jsonargparse_fork_{fn_name}_{action.dest}_typehint")
    fn = '%(fn_name)s(){ %(fn_typehint)s "%(choices)s" "$1" "%(message)s"; }' % {
        "fn_name": fn_name,
        "fn_typehint": fn_typehint,
        "choices": " ".join(choices),
        "message": message,
    }
    shtab_preambles.get().append(fn)
    action.complete = {"bash": fn_name}


def get_typehint_choices(typehint, prefix, parser, skip, choices=None, added_subclasses=None) -> List[str]:
    if choices is None:
        choices = []
    if not added_subclasses:
        added_subclasses = set()
    if typehint is bool:
        choices.extend(["true", "false"])
    elif typehint is type(None):
        choices.append("null")
    elif is_subclass(typehint, Enum):
        choices.extend(list(typehint.__members__))
    else:
        origin = get_typehint_origin(typehint)
        if origin == Union:
            for subtype in typehint.__args__:
                if subtype in added_subclasses:
                    continue
                get_typehint_choices(subtype, prefix, parser, skip, choices, added_subclasses)
        elif ActionTypeHint.is_subclass_typehint(typehint):
            added_subclasses.add(typehint)
            choices.extend(add_subactions_and_get_subclass_choices(typehint, prefix, parser, skip, added_subclasses))
        elif origin in callable_origin_types:
            return_type = get_callable_return_type(typehint)
            if return_type and ActionTypeHint.is_subclass_typehint(return_type):
                num_args = len(typehint.__args__) - 1
                skip.add(num_args)
                choices.extend(
                    add_subactions_and_get_subclass_choices(return_type, prefix, parser, skip, added_subclasses)
                )

    return [] if choices == ["null"] else choices


def add_subactions_and_get_subclass_choices(typehint, prefix, parser, skip, added_subclasses) -> List[str]:
    choices = []
    paths = get_all_subclass_paths(typehint)
    init_args = defaultdict(list)
    subclasses = defaultdict(list)
    for path in paths:
        choices.append(path)
        cls = import_object(path)
        params = get_signature_parameters(cls)
        num_skip = next((s for s in skip if isinstance(s, int)), 0)
        if num_skip > 0:
            params = params[num_skip:]
        for param in params:
            if param.name not in skip:
                init_args[param.name].append(param.annotation)
                subclasses[param.name].append(path.rsplit(".", 1)[-1])

    for name, subtypes in init_args.items():
        option_string = f"{prefix}.{name}"
        if option_string not in parser._option_string_actions:
            action = parser.add_argument(option_string)
            for subtype in unique(subtypes):
                subchoices = get_typehint_choices(subtype, option_string, parser, skip, None, added_subclasses)
                if shtab_shell.get() == "bash":
                    message = f"Expected type: {type_to_str(subtype)}; "
                    message += f"Accepted by subclasses: {', '.join(subclasses[name])}"
                    add_bash_typehint_completion(parser, action, message, subchoices)
                elif subchoices:
                    action.choices = subchoices

    return choices


def get_help_class_choices(typehint) -> List[str]:
    choices = []
    if get_typehint_origin(typehint) == Union:
        for subtype in typehint.__args__:
            if inspect.isclass(subtype):
                choices.extend(get_help_class_choices(subtype))
    else:
        choices = get_all_subclass_paths(typehint)
    return choices



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains a post-processor builder used in the DeepLab model."""

import tensorflow as tf

from deeplab2 import common
from deeplab2 import config_pb2
from deeplab2.data import dataset
from deeplab2.model import utils
from deeplab2.model.post_processor import max_deeplab
from deeplab2.model.post_processor import panoptic_deeplab


def get_post_processor(
        config: config_pb2.ExperimentOptions,
        dataset_descriptor: dataset.DatasetDescriptor) -> tf.keras.layers.Layer:
    """Initializes a DeepLab post-processor.

    Args:
      config: A config_pb2.ExperimentOptions configuration.
      dataset_descriptor: A dataset.DatasetDescriptor.

    Returns:
      PostProcessor: A post-processor depending on the configuration.
    """
    supported_tasks = utils.get_supported_tasks(config)
    if config.model_options.WhichOneof('meta_architecture') == 'max_deeplab':
        return max_deeplab.PostProcessor(config, dataset_descriptor)
    if common.TASK_PANOPTIC_SEGMENTATION in supported_tasks:
        return panoptic_deeplab.PostProcessor(config, dataset_descriptor)
    return panoptic_deeplab.SemanticOnlyPostProcessor()



"""
A package that holds third-party libraries that some geowatch tasks require.

To regenerate static submodule:

    # Checkout submodules

    # Run script:
    python ~/code/geowatch/geowatch_tpl/snapshot_submodules.py
"""

import os
import sys
# Adds the "modules" subdirectory to the python path.
# See https://gitlab.kitware.com/smart/geowatch/-/merge_requests/148#note_1050127
# for discussion of how to refactor this in the future.


# By default a few modules get added, but we should port these to the
# submodule_static repo and require they be handled on an as-needed basis.
TPL_DPATH = os.path.dirname(__file__)
MODULE_DPATH = os.path.join(os.path.dirname(__file__), 'modules')
sys.path.append(MODULE_DPATH)


FORCE_STATIC = int(os.environ.get('GEOWATCH_STATIC_TPL', '0'))


STATIC_SUBMODULES = {
    'scalemae': {
        'rel_dpath': 'scale-mae/scalemae',
        'ignore': [
            'scale-mae/scalemae/splits',
        ]
    },
    'torchview': {
        'rel_dpath': 'torchview/torchview',
    },
    'lop':
    {
        'rel_dpath': 'loss-of-plasticity/lop',
        'parts': [
            'loss-of-plasticity/lop/__init__.py',
            'loss-of-plasticity/lop/algos',
        ]
    },
    'segment_anything': {
        'rel_dpath': 'segment-anything/segment_anything',
        'parts': [
            'segment-anything/segment_anything',
            'segment-anything/LICENSE',
        ]
    },
    'jsonargparse_fork': {
        'rel_dpath': 'jsonargparse/jsonargparse_fork',
        'parts': [
            'jsonargparse/jsonargparse_fork',
            'jsonargparse/LICENSE.rst',
        ]
    },
    'detectron2': {
        'rel_dpath': 'detectron2/detectron2',
        'parts': [
            'detectron2/detectron2',
            'detectron2/LICENSE',
        ]
    },
}


def import_submodule(submod_name):
    """
    If the developer version of the submodule exists, use that.
    Otherwise use the static version.

    Args:
        submod_name (str): registered TPL module name

    Returns:
        ModuleType

    CommandLine:
        FORCE_STATIC=1 xdoctest -m geowatch_tpl import_submodule
        FORCE_STATIC=0 xdoctest -m geowatch_tpl import_submodule

    Example:
        >>> import geowatch_tpl
        >>> import ubelt as ub
        >>> for key in geowatch_tpl.STATIC_SUBMODULES:
        >>>     module = geowatch_tpl.import_submodule(key)
        >>>     print('{} module = {}'.format(key, ub.urepr(module, nl=1)))
    """
    import ubelt as ub
    if submod_name in sys.modules:
        return sys.modules[submod_name]

    tpl_dpath = ub.Path(TPL_DPATH)

    dev_mod_dpath = tpl_dpath / 'modules'
    dev_submod_dpath = tpl_dpath / 'submodules'
    static_submod_dpath = tpl_dpath / 'submodules_static'

    if submod_name in STATIC_SUBMODULES:
        info = STATIC_SUBMODULES[submod_name]
        rel_dpath = info['rel_dpath']
        cand1 = dev_submod_dpath / rel_dpath
        cand2 = static_submod_dpath / rel_dpath
        if cand1.exists() and not FORCE_STATIC:
            new_module_dpath = cand1
        else:
            assert cand2 is not None and cand2.exists()
            new_module_dpath = cand2
    else:
        # Assume we have a submodule with the same repo name if
        # it is unregistered here.
        import warnings
        cand_old = dev_mod_dpath / submod_name
        cand_new = dev_submod_dpath / submod_name / submod_name
        if cand_old.exists():
            warnings.warn('Warning: Unregistered submodule (old style)')
            new_module_dpath = cand_old
        else:
            warnings.warn('Warning: Unregistered submodule (new style)')
            new_module_dpath = cand_new
        assert new_module_dpath.exists()

    new_sys_dpath = os.fspath(new_module_dpath.parent)
    # sys.path.append(new_sys_dpath)
    sys.path.insert(0, new_sys_dpath)
    module = ub.import_module_from_name(submod_name)
    return module





def test_multimodel_with_null_sensor():
    """
    Test that unspecified sensors (i.e. None Values) work well enough and are
    translated to "*".
    """
    import numpy as np
    kwargs = {
        'name': 'unnamed_model',
        'weight_decay': 0.0,
        'lr_scheduler': 'CosineAnnealingLR',
        'class_weights': 'auto',
        'saliency_weights': 'auto',
        'stream_channels': 16,
        'tokenizer': 'rearrange',
        'token_norm': 'none',
        'arch_name': 'smt_it_stm_p2',
        'decoder': 'mlp',
        'dropout': 0.1,
        'window_size': 4,
    }
    dataset_stats = {
        'unique_sensor_modes': {
            ('*', 'gray'),
        },
        'sensor_mode_hist': {
            (None, 'gray'): 10000,
        },
        'input_stats': {
            (None, 'gray'): {
                'mean': np.array([[[0.5]]], dtype=np.float64),
                'std': np.array([[[0.5]]], dtype=np.float64),
                'min': np.array([[[0.]]], dtype=np.float64),
                'max': np.array([[[1.]]], dtype=np.float64),
                'n': np.array([[[100]]], dtype=np.float64)
            }
        },
        'class_freq': {
            'nevative': 100000,
            'positive': 10000,
        },
        'modality_input_stats': {
            (None, 'gray', None): {
                'mean': np.array([0.5], dtype=np.float64),
                'std': np.array([0.5], dtype=np.float64),
                'min': np.array([0.], dtype=np.float64),
                'max': np.array([1.], dtype=np.float64),
                'n': np.array([100], dtype=np.float64)
            },
        },
        'annot_class_freq': {},
        'track_class_freq': {},
    }
    input_sensorchan = None
    input_channels = None
    classes = ['nevative', 'positive']

    from geowatch.tasks.fusion.methods.channelwise_transformer import MultimodalTransformer
    self = MultimodalTransformer(classes=classes, dataset_stats=dataset_stats,
                                 input_sensorchan=input_sensorchan,
                                 input_channels=input_channels, **kwargs)
    assert list(self.input_norms.keys())  == ['*']
    assert list(self.input_norms['*'].keys())  == ['gray']



import random

from .functional import to_tensor, center_crop

__all__ = ["CenterCrop", "ToTensor", "RandomBool", "ConditionApply"]


class CenterCrop(object):
    def __init__(self, output_size):
        self.output_size = output_size

    def __call__(self, tensor):
        return center_crop(tensor, self.output_size)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class ToTensor(object):
    """Convert a ``PIL Image`` or ``numpy.ndarray`` to tensor without typecasting and rescaling.
    Converts a PIL Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.ByteTensor of shape (C x H x W) in the range [0, 255]
    if the PIL Image belongs to one of the modes (L, LA, P, I, F, RGB, YCbCr, RGBA, CMYK, 1)
    or if the numpy.ndarray has dtype = np.uint8
    """

    def __call__(self, pic):
        """
        Args:
            pic (PIL Image or numpy.ndarray): Image to be converted to tensor.
        Returns:
            Tensor: Converted image.
        """
        return to_tensor(pic)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class RandomBool(object):
    """Produce a random boolean with p probability for it to be True

    Args:
        p (float): probability
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self):
        return random.random() < self.p


class ConditionApply(object):
    """Apply a transformation if condition is met

    Args:
        transform:
    """

    def __init__(self, transform):
        self.transform = transform

    def __call__(self, tensor, condition):
        """

        :param tensor:
        :param condition (bool): True: apply, False: do not apply
        :return:
        """
        if condition:
            tensor = self.transform(tensor)
        return tensor

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        format_string += '\n'
        format_string += '    {0}'.format(self.transform)
        format_string += '\n)'
        return format_string



"""
This file defines a ProcessNode for each CLI process in a pipeline.
It is important that each CLI have well defined input and output paths.
"""
from geowatch.mlops.pipeline_nodes import ProcessNode
from geowatch.mlops.pipeline_nodes import PipelineDAG
import ubelt as ub

# Normally we want to invoke installed Python modules so we can abstract away
# hard coded paths, but for this example we will avoid that for simplicity.
try:
    EXAMPLE_DPATH = ub.Path(__file__).parent
except NameError:
    # for developer convinience
    EXAMPLE_DPATH = ub.Path('~/code/geowatch/docs/source/manual/tutorial/examples/mlops/mlops_example_module').expanduser()


class Stage1_Predict(ProcessNode):
    """
    Example:
        >>> import sys, ubelt
        >>> sys.path.append(ubelt.expandpath('~/code/geowatch/docs/source/manual/tutorial/examples/mlops'))
        >>> from mlops_example_module.pipelines import *  # NOQA
        >>> self = Stage1_Predict()
        >>> print(self.command)
    """
    name = 'stage1_predict'
    executable = f'python {EXAMPLE_DPATH}/cli/stage1_predict.py'

    in_paths = {
        'src_fpath',
    }
    out_paths = {
        'dst_fpath': 'stage1_prediction.json',
        'dst_dpath': '.',
    }
    primary_out_key = 'dst_fpath'

    algo_params = {
        'param1': 1,
    }
    perf_params = {
        'workers': 0,
    }

    def load_result(self, node_dpath):
        import json
        from geowatch.mlops.aggregate_loader import new_process_context_parser
        from geowatch.utils import util_dotdict
        output_fpath = node_dpath / self.out_paths[self.primary_out_key]
        result = json.loads(output_fpath.read_text())
        proc_item = result['info'][-1]
        nest_resolved = new_process_context_parser(proc_item)
        flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
        flat_resolved = flat_resolved.insert_prefix(self.name, index=1)
        return flat_resolved


class Stage1_Evaluate(ProcessNode):
    name = 'stage1_evaluate'
    executable = f'python {EXAMPLE_DPATH}/cli/stage1_evaluate.py'

    in_paths = {
        'true_fpath',
        'pred_fpath',
    }
    out_paths = {
        'out_fpath': 'stage1_evaluation.json',
    }
    algo_params = {
    }
    perf_params = {
        'workers': 0,
    }

    def load_result(self, node_dpath):
        """
        The specific implementation uses convinience functions that rely on how
        the script implemention stores results, but any manual implementation
        will work if it returns a flat dict items of the form:
        ``"metrics.<node_name>.<metric>": <value>``.

        Returns:
            Dict[str, Any]
        """
        import json
        from geowatch.mlops.aggregate_loader import new_process_context_parser
        from geowatch.utils import util_dotdict
        output_fpath = node_dpath / self.out_paths[self.primary_out_key]
        result = json.loads(output_fpath.read_text())
        proc_item = result['info'][-1]
        nest_resolved = new_process_context_parser(proc_item)
        nest_resolved['metrics'] = result['result']
        flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
        flat_resolved = flat_resolved.insert_prefix(self.name, index=1)
        return flat_resolved

    def default_metrics(self):
        """
        Returns:
            List[Dict]: containing information on how to interpret and
            prioritize the metrics returned here.
        """
        metric_infos = [
            {
                'metric': 'accuracy',
                'objective': 'maximize',
                'primary': True,
            },
            {
                'metric': 'hamming_distance',
                'objective': 'minimize',
                'primary': True,
            }
        ]
        return metric_infos

    @property
    def default_vantage_points(self):
        vantage_points = [
            {
                'metric1': 'metrics.stage1_evaluate.accuracy',
                'metric2': 'metrics.stage1_evaluate.hamming_distance',
            },
        ]
        return vantage_points


def my_demo_pipeline():
    """
    Example:
        >>> from mlops_example_module.pipelines import *  # NOQA
        >>> dag = my_demo_pipeline()
        >>> dag.configure({
        ...     'stage1_predict.src_fpath': 'my-input-path',
        ... })
        >>> dag.print_graphs(shrink_labels=1, show_types=1)
        >>> queue = dag.make_queue()['queue']
        >>> queue.print_commands(with_locks=0)

    Ignore:
        from graphid import util
        proc_graph = dag.proc_graph.copy()
        util.util_graphviz.dump_nx_ondisk(proc_graph, 'proc_graph.png')
        import xdev
        xdev.startfile('proc_graph.png')
    """
    # Define the nodes as stages in the pipeline
    nodes = {}
    nodes['stage1_predict'] = Stage1_Predict()
    nodes['stage1_evaluate'] = Stage1_Evaluate()

    # Next we build the edges

    # Outputs can be connected to inputs
    nodes['stage1_predict'].outputs['dst_fpath'].connect(nodes['stage1_evaluate'].inputs['pred_fpath'])

    # Inputs can be connected to other inputs if they are reused.
    nodes['stage1_predict'].inputs['src_fpath'].connect(nodes['stage1_evaluate'].inputs['true_fpath'])

    dag = PipelineDAG(nodes)
    dag.build_nx_graphs()
    return dag



#!/usr/bin/env python3
"""
TODO:
    - [ ] Add relevant documentation to point at
          ~/code/geowatch/geowatch/mlops/repackager.py so it is easy for the user to
          unpack / repack models.

"""
import scriptconfig as scfg
import ubelt as ub


class UnpackModelCLI(scfg.DataConfig):
    """
    Unpack the core components of a torch package to make them suitable for
    environment-agnostic use or repackaging.

    SeeAlso
    -------
    python -m geowatch/mlops/repackager
    """
    fpath = scfg.Value(None, help='the path to a torch package', position=1)
    dst_dpath = scfg.Value(None, help='Path to a destination directory to write to. If unspecfied chooses one.')

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> from geowatch.cli.experimental.unpack_model import *  # NOQA
            >>> cmdline = 0
            >>> kwargs = dict()
            >>> kwargs['fpath'] = '/data/joncrall/dvc-repos/smart_phase3_expt/models/fusion/Drop8-ARA-Cropped2GSD-V1/packages/Drop8-ARA-Cropped2GSD-V1_allsensors_V001/Drop8-ARA-Cropped2GSD-V1_allsensors_V001_epoch0_step21021.pt'
            >>> cls = UnpackModelCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        import rich
        from rich.markup import escape
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        rich.print('config = ' + escape(ub.urepr(config, nl=1)))
        package_fpath = config.fpath
        unpack_model(package_fpath, dst_dpath=config.dst_dpath)


def unpack_model(package_fpath, dst_dpath=None):
    """
    Extracts and writes extracted files to disk.

    Returns:
        Dict[str, ub.Path]: mapping from keys to written paths on disk

    Ignore:
        package_fpath = '/data/joncrall/dvc-repos/smart_phase3_expt/models/fusion/Drop8-ARA-Cropped2GSD-V1/packages/Drop8-ARA-Cropped2GSD-V1_allsensors_V001/Drop8-ARA-Cropped2GSD-V1_allsensors_V001_epoch0_step21021.pt'
        result = unpack_model(package_fpath)
        checkpoint_fpath = result['ckpt_fpath']
        from geowatch.mlops.repackager import repackage
        new_package_fpath = repackage(checkpoint_fpath=checkpoint_fpath)[0]
        round2_result = unpack_model(new_package_fpath)

        import kwutil
        config1 = kwutil.Yaml.load(result['config_fpath'])
        config2 = kwutil.Yaml.load(round2_result['config_fpath'])
        config1 == config2

        from kwcoco.util.util_json import indexable_diff
        info = indexable_diff(config1, config2)
        assert info['similarity'] == 1.0

    Example:
        >>> import ubelt as ub
        >>> from geowatch.tasks.fusion.methods.channelwise_transformer import *  # NOQA
        >>> dpath = ub.Path.appdir('geowatch/tests/package').ensuredir()
        >>> package_fpath = dpath / 'my_package.pt'
        >>> dst_dpath = dpath / 'unpacked_packages'
        >>> if not package_fpath.exists():
        >>>     # Use one of our fusion.architectures in a test
        >>>     from geowatch.tasks.fusion import methods
        >>>     from geowatch.tasks.fusion import datamodules
        >>>     model = self = methods.MultimodalTransformer(
        >>>         arch_name="smt_it_joint_p2", input_sensorchan=5,
        >>>         change_head_hidden=0, saliency_head_hidden=0,
        >>>         class_head_hidden=0)
        >>>     # Save the model (TODO: need to save datamodule as well)
        >>>     model.save_package(package_fpath)
        >>> results = unpack_model(package_fpath)
        >>> # Test repackage
        >>> from geowatch.mlops.repackager import repackage_single_checkpoint
        >>> repackage_single_checkpoint
    """
    from kwutil.util_yaml import Yaml
    import torch

    package_fpath = ub.Path(package_fpath)
    package_content = extract_package_contents(package_fpath)

    if dst_dpath is None:
        dst_dpath = package_fpath.parent / package_fpath.stem
    else:
        dst_dpath = ub.Path(dst_dpath)
    dst_dpath.ensuredir()

    config_fpath = dst_dpath / 'config.yaml'
    config_fpath.write_text(Yaml.dumps(package_content['config']))

    ckpt_name = package_fpath.stem + '.ckpt'
    ckpt_dpath = (dst_dpath / 'checkpoints').ensuredir()
    checkpoint_fpath = ckpt_dpath / ckpt_name

    with open(checkpoint_fpath, 'wb') as file:
        torch.save(package_content['checkpoint'], file)

    result = {}
    result['dpath'] = dst_dpath
    result['ckpt_fpath'] = checkpoint_fpath
    result['config_fpath'] = config_fpath
    return result


def extract_package_contents(package_fpath):
    """
    Returns:
        dict
    """
    from geowatch.tasks.fusion import utils
    # from geowatch.utils import util_netharn
    from geowatch.monkey import monkey_torchmetrics
    monkey_torchmetrics.fix_torchmetrics_compatability()

    package_fpath = ub.Path(package_fpath)

    if not package_fpath.exists():
        if package_fpath.augment(tail='.dvc').exists():
            raise Exception('model does not exist, but its dvc file does')
        else:
            raise Exception('model does not exist')

    # file_stat = package_fpath.stat()
    # TODO: generalize the load-package

    # TODO: is it possible to extract only the weights or only the torch /
    # numpy objects from a pickle file and ignore any part that references an
    # unavailable package?
    raw_module = utils.load_model_from_package(package_fpath)
    if hasattr(raw_module, 'module'):
        module = raw_module.module
    else:
        module = raw_module

    # TODO: get the category freq

    module.__class__
    module.hparams

    HACK_REMOVE_PROBLEMATIC_INFO = 1
    if HACK_REMOVE_PROBLEMATIC_INFO:
        try:
            raw_module.hparams['dataset_stats'].pop('modality_input_stats', None)
        except AttributeError:
            ...

    # Construct a checkpoint the repackager will accept.
    checkpoint = {}
    checkpoint['state_dict'] = raw_module.state_dict()
    try:
        # TODO: remove numpy arrays if possible
        # They can cause read errors if there are binary incompatabilities
        # todo: make configurable?
        import kwutil
        checkpoint['hyper_parameters'] = kwutil.Json.ensure_serializable(raw_module.hparams)
    except Exception:
        checkpoint['hyper_parameters'] = raw_module.hparams

    package_header = utils.load_model_header(package_fpath)
    config_content = package_header.get('config', None)

    if HACK_REMOVE_PROBLEMATIC_INFO:
        if config_content and 'model' in config_content:
            if 'class_path' in config_content['model']:
                if config_content['model']['class_path'].startswith('watch.'):
                    config_content['model']['class_path'] = 'geo' + config_content['model']['class_path']

    package_content = {}
    package_content['config'] = config_content
    package_content['checkpoint'] = checkpoint
    return package_content


# def unpack_model_backup(package_fpath):
#     import zipfile
#     zfile = zipfile.ZipFile(package_fpath, 'r')
#     names = zfile.namelist()

#     cand_model_pkl = [n for n in names if n.endswith('model.pkl')]
#     assert len(cand_model_pkl) == 1
#     model_pkl = cand_model_pkl[0]
#     file = zfile.open(model_pkl)

#     data = file.read()
#     import pickle
#     loaded = pickle.loads(data)


__cli__ = UnpackModelCLI

if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/geowatch/geowatch/cli/experimental/unpack_model.py
        python -m geowatch.cli.experimental.unpack_model
    """
    __cli__.main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains common methods and constants used across this framework."""

# Prediction keys used by the model output dictionary.
PRED_PANOPTIC_KEY = 'panoptic_pred'
PRED_SEMANTIC_KEY = 'semantic_pred'
PRED_INSTANCE_KEY = 'instance_pred'
PRED_INSTANCE_CENTER_KEY = 'instance_center_pred'

PRED_SEMANTIC_LOGITS_KEY = 'semantic_logits'
PRED_SEMANTIC_PROBS_KEY = 'semantic_probs'
PRED_SEMANTIC_SCORES_KEY = 'semantic_scores'
PRED_INSTANCE_SCORES_KEY = 'instance_scores'
PRED_CENTER_HEATMAP_KEY = 'center_heatmap'
PRED_OFFSET_MAP_KEY = 'offset_map'
PRED_FRAME_OFFSET_MAP_KEY = 'frame_offset_map'
PRED_NEXT_OFFSET_MAP_KEY = 'next_offset_map'
PRED_NEXT_PANOPTIC_KEY = 'next_panoptic_pred'
PRED_CONCAT_NEXT_PANOPTIC_KEY = 'concat_next_panoptic_pred'
PRED_DEPTH_KEY = 'depth_pred'

PRED_PIXEL_SPACE_NORMALIZED_FEATURE_KEY = 'pixel_space_normalized_feature'
PRED_PIXEL_SPACE_MASK_LOGITS_KEY = 'pixel_space_mask_logits'
PRED_TRANSFORMER_CLASS_LOGITS_KEY = 'transformer_class_logits'

PRED_AUXILIARY_OUTPUTS = 'auxiliary_outputs_pred'

# Ground-truth keys used by the model.
GT_PANOPTIC_KEY = 'panoptic_gt'
GT_SEMANTIC_KEY = 'semantic_gt'
GT_INSTANCE_CENTER_KEY = 'instance_center_gt'
GT_INSTANCE_REGRESSION_KEY = 'instance_regression_gt'
GT_FRAME_OFFSET_KEY = 'frame_offset_gt'
GT_IS_CROWD = 'is_crowd_gt'
GT_THING_ID_MASK_KEY = 'thing_id_mask_gt'
GT_THING_ID_CLASS_KEY = 'thing_id_class_gt'
GT_NEXT_INSTANCE_REGRESSION_KEY = 'next_instance_regression_gt'
GT_DEPTH_KEY = 'depth_gt'

# Raw labels.
GT_PANOPTIC_RAW = 'panoptic_raw'
GT_SEMANTIC_RAW = 'semantic_raw'
GT_IS_CROWD_RAW = 'is_crowd_raw'
GT_SIZE_RAW = 'size_raw'
GT_NEXT_PANOPTIC_RAW = 'next_panoptic_raw'
GT_DEPTH_RAW = 'depth_raw'

# Loss keys.
SEMANTIC_LOSS = 'semantic_loss'
CENTER_LOSS = 'center_loss'
REGRESSION_LOSS = 'regression_loss'
MOTION_LOSS = 'motion_loss'
NEXT_REGRESSION_LOSS = 'next_regression_loss'
DEPTH_LOSS = 'depth_loss'
PQ_STYLE_LOSS = 'pq_style_loss'
# The PQ-style loss consists of a class term and a mask dice term.
PQ_STYLE_LOSS_CLASS_TERM = 'pq_style_loss_class_term'
PQ_STYLE_LOSS_MASK_DICE_TERM = 'pq_style_loss_mask_dice_term'
MASK_ID_CROSS_ENTROPY_LOSS = 'mask_id_cross_entropy_loss'
INSTANCE_DISCRIMINATION_LOSS = 'instance_discrimination_loss'
TOTAL_LOSS = 'total_loss'

# Weight keys used by the model.
SEMANTIC_LOSS_WEIGHT_KEY = 'semantic_loss_weight'
CENTER_LOSS_WEIGHT_KEY = 'center_loss_weight'
REGRESSION_LOSS_WEIGHT_KEY = 'regression_loss_weight'
FRAME_REGRESSION_LOSS_WEIGHT_KEY = 'frame_regression_loss_weight'
NEXT_REGRESSION_LOSS_WEIGHT_KEY = 'next_regression_loss_weight'
DEPTH_LOSS_WEIGHT_KEY = 'depth_loss_weight'

# Misc.
RESIZED_IMAGE = 'resized_image'
IMAGE = 'image'
IMAGE_NAME = 'image_name'
# For PVPS, we use sequence_id + frame_id to uniquely identify a frame.
SEQUENCE_ID = 'sequence_id'
FRAME_ID = 'frame_id'
NEXT_IMAGE = 'next_image'

# TfExample keys.
KEY_ENCODED_IMAGE = 'image/encoded'
KEY_ENCODED_PREV_IMAGE = 'prev_image/encoded'
KEY_ENCODED_NEXT_IMAGE = 'next_image/encoded'
KEY_IMAGE_FILENAME = 'image/filename'
KEY_IMAGE_FORMAT = 'image/format'
KEY_IMAGE_HEIGHT = 'image/height'
KEY_IMAGE_WIDTH = 'image/width'
KEY_IMAGE_CHANNELS = 'image/channels'
KEY_ENCODED_LABEL = 'image/segmentation/class/encoded'
KEY_ENCODED_PREV_LABEL = 'prev_image/segmentation/class/encoded'
KEY_ENCODED_NEXT_LABEL = 'next_image/segmentation/class/encoded'
KEY_LABEL_FORMAT = 'image/segmentation/class/format'
KEY_SEQUENCE_ID = 'video/sequence_id'
KEY_FRAME_ID = 'video/frame_id'
KEY_ENCODED_DEPTH = 'image/depth/encoded'
KEY_DEPTH_FORMAT = 'image/depth/format'

# TfExample keys for the panoramic setting, where individual camera keys are
# obtained by formatting with the camera name.
KEY_PER_CAMERA_ENCODED_IMAGE = 'image/encoded/%s'
KEY_PER_CAMERA_ENCODED_NEXT_IMAGE = 'next_image/encoded/%s'
KEY_PER_CAMERA_IMAGE_HEIGHT = 'image/height/%s'
KEY_PER_CAMERA_IMAGE_WIDTH = 'image/width/%s'
KEY_PER_CAMERA_ENCODED_LABEL = 'image/segmentation/class/encoded/%s'
KEY_PER_CAMERA_ENCODED_NEXT_LABEL = 'next_image/segmentation/class/encoded/%s'
KEY_PER_CAMERA_ENCODED_DEPTH = 'image/depth/encoded/%s'

# Checkpoint Items
# All models
CKPT_SEMANTIC_LAST_LAYER = 'semantic_last_layer'

# DeepLabV3
CKPT_DEEPLABV3_ASPP = 'deeplab_v3_aspp'
CKPT_DEEPLABV3_CLASSIFIER_CONV_BN_ACT = 'classifier_conv_bn_act'

# DeepLabV3+
CKPT_DEEPLABV3PLUS_ASPP = 'deeplab_v3plus_aspp'
CKPT_DEEPLABV3PLUS_PROJECT_CONV_BN_ACT = 'deeplab_v3plus_project_conv_bn_act'
CKPT_DEEPLABV3PLUS_FUSE = 'deeplab_v3plus_fuse'

# Panoptic-DeepLab
CKPT_SEMANTIC_DECODER = 'semantic_decoder'
CKPT_SEMANTIC_HEAD_WITHOUT_LAST_LAYER = 'semantic_head_without_last_layer'

CKPT_INSTANCE_DECODER = 'instance_decoder'
CKPT_INSTANCE_CENTER_HEAD_WITHOUT_LAST_LAYER = ('instance_center_head'
                                                '_without_last_layer')
CKPT_INSTANCE_CENTER_HEAD_LAST_LAYER = 'instance_center_head_last_layer'
CKPT_INSTANCE_REGRESSION_HEAD_WITHOUT_LAST_LAYER = ('instance_regression_head'
                                                    '_without_last_layer')
CKPT_INSTANCE_REGRESSION_HEAD_LAST_LAYER = 'instance_regression_head_last_layer'

# Motion-DeepLab
CKPT_MOTION_REGRESSION_HEAD_WITHOUT_LAST_LAYER = ('motion_regression_head'
                                                  '_without_last_layer')
CKPT_MOTION_REGRESSION_HEAD_LAST_LAYER = 'motion_regression_head_last_layer'

# ViP-DeepLab
CKPT_NEXT_INSTANCE_DECODER = 'next_instance_decoder'
CKPT_NEXT_INSTANCE_REGRESSION_HEAD_WITHOUT_LAST_LAYER = (
    'next_instance_regression_head_without_last_layer')
CKPT_NEXT_INSTANCE_REGRESSION_HEAD_LAST_LAYER = (
    'next_instance_regression_head_last_layer')
CKPT_DEPTH_HEAD_WITHOUT_LAST_LAYER = 'depth_head_without_last_layer'
CKPT_DEPTH_HEAD_LAST_LAYER = 'depth_head_last_layer'

# MaX-DeepLab
CKPT_PIXEL_SPACE_HEAD = 'pixel_space_head'
CKPT_TRANSFORMER_MASK_HEAD = 'transformer_mask_head'
CKPT_TRANSFORMER_CLASS_HEAD = 'transformer_class_head'
CKPT_PIXEL_SPACE_FEATURE_BATCH_NORM = 'pixel_space_feature_batch_norm'
CKPT_PIXEL_SPACE_MASK_BATCH_NORM = 'pixel_space_mask_batch_norm'

# Supported Tasks
TASK_PANOPTIC_SEGMENTATION = 'panoptic_segmentation'
TASK_INSTANCE_SEGMENTATION = 'instance_segmentation'
TASK_VIDEO_PANOPTIC_SEGMENTATION = 'video_panoptic_segmentation'
TASK_DEPTH_AWARE_VIDEO_PANOPTIC_SEGMENTATION = (
    'depth_aware_video_panoptic_segmentation')



from collections import OrderedDict

import PIL
import numpy as np
import torch
import torchvision

import kornia

import torch_lydorn.kornia
import torch_lydorn.torchvision

from lydorn_utils import print_utils


class Print(object):
    """Convert polygons to a single graph"""

    def __init__(self):
        pass

    def __call__(self, sample):
        print("\n")
        print(sample.keys())
        for key, item in sample.items():
            if isinstance(item, np.ndarray) or isinstance(item, torch.Tensor):
                if len(item.shape):
                    print(key, type(item), item.shape, item.dtype, item.min(), item.max())
                else:
                    print(key, type(item), item, item.dtype, item.min(), item.max())
            elif isinstance(item, PIL.Image.Image):
                print(key, type(item), item.size, item.mode, np.array(item).min(), np.array(item).max())
            elif isinstance(item, list):
                print(key, type(item[0]), len(item))
        # exit()
        # print(sample["image"].dtype)
        # print(sample["image"].shape)
        # print(sample["image"].min())
        # print(sample["image"].max())
        # for key, value in sample.items():
        #     print(key + ":")
        #     if type(value) == np.ndarray:
        #         print(value.shape)
        #     elif type(value) == list:
        #         print(len(value))
        #     else:
        #         print("a")
        return sample


class CudaDataAugmentation(object):
    def __init__(self, input_patch_size: int, vflip: bool, affine: bool, scaling: list, color_jitter: bool):
        self.input_patch_size = input_patch_size
        self.vflip = vflip
        self.affine = affine
        self.scaling = scaling
        self.color_jitter = None
        if color_jitter:
            self.color_jitter = kornia.augmentation.ColorJitter(brightness=0.05, contrast=0.05, saturation=.5, hue=.1)
        self.tensor_keys_bilinear = ["image", "gt_polygons_image", "distances",
                                     "valid_mask"]  # Affine transform applied with bilinear sampling
        self.tensor_keys_nearest = ["sizes", "gt_crossfield_angle"]  # Affine transform applied with nearest sampling

    @staticmethod
    def get_slices(batch, keys, last_slice_stop=0):
        slices = OrderedDict()
        for key in keys:
            s = slice(last_slice_stop, last_slice_stop + batch[key].shape[1])
            last_slice_stop += batch[key].shape[1]
            slices[key] = s
        return slices

    def __call__(self, batch):
        with torch.no_grad():
            batch_size, im_channels, height, width = batch["image"].shape
            device = batch["image"].device
            batch["valid_mask"] = torch.ones((batch_size, 1, height, width), dtype=torch.float,
                                             device=device)  # Apply losses only when valid_mask is True (pad with 0 when rotating)

            # Combine all images into one for faster/easier processing (store slices to separate them later on):
            tensor_keys_bilinear = [key for key in self.tensor_keys_bilinear if key in batch]
            tensor_keys_nearest = [key for key in self.tensor_keys_nearest if key in batch]
            tensor_keys = tensor_keys_bilinear + tensor_keys_nearest
            combined = torch.cat([batch[tensor_key] for tensor_key in tensor_keys], dim=1)
            slices_bilinear = self.get_slices(batch, tensor_keys_bilinear, last_slice_stop=0)
            slices_nearest = self.get_slices(batch, tensor_keys_nearest,
                                             last_slice_stop=slices_bilinear[tensor_keys_bilinear[-1]].stop)
            bilinear_slice = slice(slices_bilinear[tensor_keys_bilinear[0]].start,
                                   slices_bilinear[tensor_keys_bilinear[-1]].stop)
            nearest_slice = slice(slices_nearest[tensor_keys_nearest[0]].start,
                                  slices_nearest[tensor_keys_nearest[-1]].stop)

            # Rotation (and translation)
            if self.affine:
                angle: torch.Tensor = torch.empty(batch_size, device=device).uniform_(-180, 180)
                # To include corner pixels if angle=45 (coords are between -1 and 1 for grid_sample):
                max_offset = np.sqrt(2) - 1
                offset: torch.Tensor = torch.empty((batch_size, 2), device=device).uniform_(-max_offset, max_offset)
                downscale_factor = None
                if self.scaling is not None:
                    downscale_factor: torch.Tensor = torch.empty(batch_size, device=device).uniform_(*self.scaling)
                affine_grid = torch_lydorn.kornia.geometry.transform.get_affine_grid(combined, angle, offset,
                                                                                     downscale_factor)
                combined[:, bilinear_slice, ...] = \
                    torch.nn.functional.grid_sample(combined[:, bilinear_slice, ...],
                                                    affine_grid, mode='bilinear')

                # Rotate sizes and anglefield with mode='nearest'
                # because it makes no sense to interpolate size values and angle values:
                combined[:, nearest_slice, ...] = torch.nn.functional.grid_sample(combined[:, nearest_slice, ...],
                                                                                  affine_grid, mode='nearest')

                # Additionally the angle field's values themselves have to be rotated:
                combined[:, slices_nearest["gt_crossfield_angle"], ...] = torch_lydorn.torchvision.transforms.functional.rotate_anglefield(
                    combined[:, slices_nearest["gt_crossfield_angle"], ...], angle)

                # The sizes and distances should be adjusted as well because of the scaling.
                if downscale_factor is not None:
                    if "sizes" in slices_nearest:
                        size_equals_one = combined[:, slices_nearest["sizes"], ...] == 1
                        combined[:, slices_nearest["sizes"], :, :] /= downscale_factor[:, None, None, None]
                        combined[:, slices_nearest["sizes"], ...][size_equals_one] = 1
                    if "distances" in slices_bilinear:
                        distance_equals_one = combined[:, slices_bilinear["distances"], ...] == 1
                        combined[:, slices_bilinear["distances"], :, :] /= downscale_factor[:, None, None, None]
                        combined[:, slices_bilinear["distances"], ...][distance_equals_one] = 1

            # Center crop
            if self.input_patch_size is not None:
                prev_image_norm = combined.shape[2] + combined.shape[3]
                combined = torch_lydorn.torchvision.transforms.functional.center_crop(combined, self.input_patch_size)
                current_image_norm = combined.shape[2] + combined.shape[3]
                # Sizes and distances are affected by this because they are relative to the image's size.
                # All non-one pixels have to be renormalized:
                size_ratio = prev_image_norm / current_image_norm
                if "sizes" in slices_nearest:
                    combined[:, slices_nearest["sizes"], ...][
                        combined[:, slices_nearest["sizes"], ...] != 1] *= size_ratio
                if "distances" in slices_bilinear:
                    combined[:, slices_bilinear["distances"], ...][
                        combined[:, slices_bilinear["distances"], ...] != 1] *= size_ratio

            # vflip
            if self.vflip:
                to_flip: torch.Tensor = torch.empty(batch_size, device=device).uniform_(0, 1) < 0.5
                combined[to_flip] = kornia.geometry.transform.vflip(combined[to_flip])
                combined[
                    to_flip, slices_nearest[
                        "gt_crossfield_angle"], ...] = torch_lydorn.torchvision.transforms.functional.vflip_anglefield(
                    combined[to_flip, slices_nearest["gt_crossfield_angle"], ...])

            # Split data:
            batch["image"] = combined[:, slices_bilinear["image"], ...]
            if "gt_polygons_image" in slices_bilinear:
                batch["gt_polygons_image"] = combined[:, slices_bilinear["gt_polygons_image"], ...]
            if "distances" in slices_bilinear:
                batch["distances"] = combined[:, slices_bilinear["distances"], ...]
            batch["valid_mask"] = 0.99 < combined[:, slices_bilinear["valid_mask"], ...]  # Take a very high threshold to remove fuzzy pixels

            if "sizes" in slices_nearest:
                batch["sizes"] = combined[:, slices_nearest["sizes"], ...]
            batch["gt_crossfield_angle"] = combined[:, slices_nearest["gt_crossfield_angle"], ...]

            # Color jitter
            if self.color_jitter is not None and batch["image"].shape[1] == 3:
                batch["image"] = self.color_jitter(batch["image"])

            # --- Zero padding of sizes and distances is not correct, they should be padded with ones:
            if self.affine:
                if "sizes" in slices_nearest:
                    batch["sizes"][~batch["valid_mask"]] = 1
                if "distances" in slices_bilinear:
                    batch["distances"][~batch["valid_mask"]] = 1

        return batch


class CudaCrop(object):
    def __init__(self, input_patch_size: int):
        self.input_patch_size = input_patch_size
        self.tensor_keys = ["image", "gt_polygons_image", "distances", "valid_mask", "sizes", "gt_crossfield_angle"]

    def __call__(self, batch):
        for tensor_key in self.tensor_keys:
            if tensor_key in batch:
                batch[tensor_key] = torch_lydorn.torchvision.transforms.functional.center_crop(batch[tensor_key],
                                                                                               self.input_patch_size)
        return batch


def get_offline_transform(config, augmentations=False, to_patches=True):
    data_patch_size = config["dataset_params"]["data_patch_size"] if augmentations else config["dataset_params"][
        "input_patch_size"]
    transform_list = [
        torch_lydorn.torchvision.transforms.Map(
            transform=torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torchvision.transforms.Compose([
                    torch_lydorn.torchvision.transforms.RemoveDoubles(epsilon=0.01),
                    torch_lydorn.torchvision.transforms.FilterPolyVertexCount(min=3),
                    torch_lydorn.torchvision.transforms.ApproximatePolygon(tolerance=0.01),
                    torch_lydorn.torchvision.transforms.FilterPolyVertexCount(min=3)
                ]), key="gt_polygons")),

        torch_lydorn.torchvision.transforms.FilterEmptyPolygons(key="gt_polygons"),
    ]
    if to_patches:
        transform_list.extend([
            torch_lydorn.torchvision.transforms.ToPatches(stride=config["dataset_params"]["input_patch_size"],
                                                          size=data_patch_size),
            torch_lydorn.torchvision.transforms.FilterEmptyPolygons(key="gt_polygons"),
        ])
    transform_list.extend([
        torch_lydorn.torchvision.transforms.Map(
            transform=torchvision.transforms.Compose([
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torch_lydorn.torchvision.transforms.Rasterize(fill=True, edges=True, vertices=True,
                                                                            line_width=4, antialiasing=True),
                    key=["image", "gt_polygons"], outkey="gt_polygons_image"),
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torch_lydorn.torchvision.transforms.AngleFieldInit(line_width=6),
                    key=["image", "gt_polygons"],
                    outkey="gt_crossfield_angle")
            ])),
    ])
    offline_transform = torchvision.transforms.Compose(transform_list)
    return offline_transform


def get_offline_transform_patch(raster: bool = True, fill: bool = True, edges: bool = True, vertices: bool = True,
                                distances: bool = True, sizes: bool = True, angle_field: bool = True):
    transform_list = []
    if raster:
        if not distances and not sizes:
            rasterize_transform = torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torch_lydorn.torchvision.transforms.Rasterize(fill=fill, edges=edges, vertices=vertices,
                                                                        line_width=4, antialiasing=True,
                                                                        return_distances=False,
                                                                        return_sizes=False),
                key=["image", "gt_polygons"], outkey="gt_polygons_image")
        elif distances and sizes:
            rasterize_transform = torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torch_lydorn.torchvision.transforms.Rasterize(fill=fill, edges=edges, vertices=vertices,
                                                                            line_width=4, antialiasing=True,
                                                                            return_distances=True,
                                                                            return_sizes=True),
                    key=["image", "gt_polygons"], outkey=["gt_polygons_image", "distances", "sizes"])
        else:
            raise NotImplementedError
        transform_list.append(rasterize_transform)
    if angle_field:
        transform_list.append(
            torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torch_lydorn.torchvision.transforms.AngleFieldInit(line_width=6),
                key=["image", "gt_polygons"],
                outkey="gt_crossfield_angle")
        )

    return torchvision.transforms.Compose(transform_list)


def get_online_cpu_transform(config, augmentations=False):
    if augmentations and config["data_aug_params"]["device"] == "cpu":
        print_utils.print_error("ERROR: CPU augmentations is not supported anymore. "
                                "Look at CudaDataAugmentation to see what additional augs would need to be implemented.")
        raise NotImplementedError
    online_transform_list = []
    # Convert to PIL images
    if not augmentations \
            or (augmentations and config["data_aug_params"]["device"] == "cpu"):
        online_transform_list.extend([
            torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.ToPILImage(),
                                                               key="image"),
            torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.ToPILImage(),
                                                               key="gt_polygons_image"),
            torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.ToPILImage(),
                                                               key="gt_crossfield_angle"),
        ])
    # Add rotation data augmentation:
    if augmentations and config["data_aug_params"]["device"] == "cpu" and \
            config["data_aug_params"]["affine"]:
        online_transform_list.extend([
            torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torch_lydorn.torchvision.transforms.SampleUniform(-180, 180),
                outkey="rand_angle"),
            torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.functional.rotate,
                                                               key=["image", "rand_angle"], outkey="image",
                                                               resample=PIL.Image.BILINEAR),
            torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.functional.rotate,
                                                               key=["gt_polygons_image", "rand_angle"],
                                                               outkey="gt_polygons_image",
                                                               resample=PIL.Image.BILINEAR),
            torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.functional.rotate,
                                                               key=["gt_crossfield_angle", "rand_angle"],
                                                               outkey="gt_crossfield_angle",
                                                               resample=PIL.Image.NEAREST),
        ])

    # Crop to final size
    if not augmentations \
            or (augmentations and config["data_aug_params"]["device"] == "cpu"):
        if "input_patch_size" in config["dataset_params"]:
            online_transform_list.extend([
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torchvision.transforms.CenterCrop(config["dataset_params"]["input_patch_size"]),
                    key="image"),
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torchvision.transforms.CenterCrop(config["dataset_params"]["input_patch_size"]),
                    key="gt_polygons_image"),
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torchvision.transforms.CenterCrop(config["dataset_params"]["input_patch_size"]),
                    key="gt_crossfield_angle"),
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torch_lydorn.torchvision.transforms.CenterCrop(
                        config["dataset_params"]["input_patch_size"]),
                    key="distances"),
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torch_lydorn.torchvision.transforms.CenterCrop(
                        config["dataset_params"]["input_patch_size"]),
                    key="sizes"),
            ])

    # Random Horizontal flip:
    if augmentations and config["data_aug_params"]["device"] == "cpu" and \
            config["data_aug_params"]["vflip"]:
        online_transform_list.extend([
            torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torch_lydorn.torchvision.transforms.RandomBool(p=0.5),
                outkey="rand_flip"),
            torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torch_lydorn.torchvision.transforms.ConditionApply(
                    transform=torchvision.transforms.functional.vflip),
                key=["image", "rand_flip"], outkey="image"),
            torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torch_lydorn.torchvision.transforms.ConditionApply(
                    transform=torchvision.transforms.functional.vflip),
                key=["gt_polygons_image", "rand_flip"], outkey="gt_polygons_image"),
            torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torch_lydorn.torchvision.transforms.ConditionApply(
                    transform=torchvision.transforms.functional.vflip),
                key=["gt_crossfield_angle", "rand_flip"], outkey="gt_crossfield_angle"),
        ])

    # Other augs:
    if augmentations and config["data_aug_params"]["device"] == "cpu" and \
            config["data_aug_params"]["color_jitter"]:
        online_transform_list.append(
            torch_lydorn.torchvision.transforms.TransformByKey(
                transform=torchvision.transforms.ColorJitter(brightness=0.05, contrast=0.05,
                                                             saturation=.5, hue=.1),
                key="image")
        )
    # Convert to PyTorch tensors:
    online_transform_list.extend([
        # Print(),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="image"),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torchvision.transforms.Lambda(torch.from_numpy),
            key="image_mean"),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torchvision.transforms.Lambda(torch.from_numpy),
            key="image_std"),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="gt_polygons_image", ignore_key_error=True),
        torch_lydorn.torchvision.transforms.TransformByKey(torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="gt_crossfield_angle", ignore_key_error=True),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="distances", ignore_key_error=True),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="sizes", ignore_key_error=True),
    ])

    online_transform_list.append(
        torch_lydorn.torchvision.transforms.RemoveKeys(keys=["gt_polygons"])
    )

    online_transform = torchvision.transforms.Compose(online_transform_list)
    return online_transform


def get_eval_online_cpu_transform():
    online_transform = torchvision.transforms.Compose([
        # Print(),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="image"),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.Lambda(
            torch.from_numpy),
            key="image_mean"),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.Lambda(
            torch.from_numpy),
            key="image_std"),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="gt_polygons_image"),
        torch_lydorn.torchvision.transforms.TransformByKey(torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="gt_crossfield_angle"),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="distances"),
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torch_lydorn.torchvision.transforms.ToTensor(),
                                                           key="sizes"),
        torch_lydorn.torchvision.transforms.RemoveKeys(keys=["gt_polygons"])
    ])
    return online_transform


def get_online_cuda_transform(config, augmentations=False):

    device_transform_list = [
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.Compose([
            torchvision.transforms.Lambda(lambda tensor: tensor.float().div(255))
        ]), key="image"),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torchvision.transforms.Lambda(lambda tensor: tensor.float().div(255)),
            key="gt_polygons_image"),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torchvision.transforms.Lambda(lambda tensor: np.pi * tensor.float().div(255)),
            key="gt_crossfield_angle"),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torchvision.transforms.Lambda(lambda tensor: tensor.float()),
            key="distances", ignore_key_error=True),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torchvision.transforms.Lambda(lambda tensor: tensor.float()),
            key="sizes", ignore_key_error=True),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torchvision.transforms.Lambda(lambda tensor: tensor.float()),
            key="class_freq", ignore_key_error=True),
    ]

    if augmentations and config["data_aug_params"]["device"] == "cpu":
        if config["data_aug_params"]["affine"]:
            # Rotate angle field
            device_transform_list.append(
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torch_lydorn.torchvision.transforms.functional.rotate_anglefield,
                    key=["gt_crossfield_angle", "rand_angle"],
                    outkey="gt_crossfield_angle"))
        if config["data_aug_params"]["vflip"]:
            device_transform_list.append(
                torch_lydorn.torchvision.transforms.TransformByKey(
                    transform=torch_lydorn.torchvision.transforms.functional.vflip_anglefield,
                    key=["gt_crossfield_angle", "rand_flip"],
                    outkey="gt_crossfield_angle"))
    if config["data_aug_params"]["device"] == "cuda":
        input_patch_size = config["dataset_params"]["input_patch_size"] if "input_patch_size" in config[
            "dataset_params"] else None  # No crop if None
        if augmentations:
            device_transform_list.append(CudaDataAugmentation(input_patch_size,
                                                              config["data_aug_params"]["vflip"],
                                                              config["data_aug_params"]["affine"],
                                                              config["data_aug_params"]["scaling"],
                                                              config["data_aug_params"]["color_jitter"]))
        elif input_patch_size is not None:
            device_transform_list.append(CudaCrop(input_patch_size))
    device_transform_list.append(
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torch_lydorn.torchvision.transforms.functional.batch_normalize,
            key=["image", "image_mean", "image_std"],
            outkey="image"), )
    device_transform = torchvision.transforms.Compose(device_transform_list)
    return device_transform


def get_eval_online_cuda_transform(config):
    device_transform_list = [
        torch_lydorn.torchvision.transforms.TransformByKey(transform=torchvision.transforms.Compose([
            torchvision.transforms.Lambda(lambda tensor: tensor.float().div(255))
        ]), key="image"),
        torch_lydorn.torchvision.transforms.TransformByKey(
            transform=torch_lydorn.torchvision.transforms.functional.batch_normalize,
            key=["image", "image_mean", "image_std"],
            outkey="image")
    ]
    device_transform = torchvision.transforms.Compose(device_transform_list)
    return device_transform



"""
Script for porting relevant parts of submodules to a static path in this repo.

Ideally users of this repo do not need to ever touch submodules, but as we
develop I do want to ensure we tie TPL code to the original source as best as
possible (and contribute back!).
"""


def main():
    import geowatch_tpl
    import ubelt as ub
    import rich

    tpl_dpath = ub.Path(geowatch_tpl.__file__).parent
    dev_submod_dpath = tpl_dpath / 'submodules'
    static_submod_dpath = tpl_dpath / 'submodules_static'
    rich.print('dev_submod_dpath = {}'.format(ub.urepr(dev_submod_dpath, nl=1)))
    rich.print('static_submod_dpath = {}'.format(ub.urepr(static_submod_dpath, nl=1)))

    SUBMODULE_INFOS = list(geowatch_tpl.STATIC_SUBMODULES.items())
    rich.print('SUBMODULE_INFOS = {}'.format(ub.urepr(SUBMODULE_INFOS, nl=1)))

    rich.print('[green]--- Copy Submodules ---\n')

    for key, info in ub.ProgIter(SUBMODULE_INFOS, desc='copy submodules', verbose=3):
        rel_dpath = info['rel_dpath']
        src_dpath = dev_submod_dpath / rel_dpath
        dst_dpath = static_submod_dpath / rel_dpath

        if src_dpath.exists():
            rich.print(f'Copy: {src_dpath} -> {dst_dpath}')
            dst_dpath.delete()
            dst_dpath.parent.ensuredir()
            if 'parts' in info:
                for p in info['parts']:
                    p_src = dev_submod_dpath / p
                    print(f'p_src={p_src}')
                    p_dst = static_submod_dpath / p
                    assert p_src.exists()
                    p_dst.parent.ensuredir()
                    p_dst.delete()
                    p_src.copy(p_dst)
                    ...
            else:
                src_dpath.copy(dst_dpath)

            if 'ignore' in info:
                # Remove parts that were ignored. (dont have a good way to
                # exclude them from copy itself)
                for p in info['ignore']:
                    ignore_path = static_submod_dpath / p
                    print(f'ignore_path={ignore_path}')
                    ignore_path.delete()

        else:
            print(f'No dev: {src_dpath}')
        rich.print('[yellow]---\n')


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/geowatch_tpl/snapshot_submodules.py
    """
    main()



import json
import tempfile
import uuid

import ubelt as ub
import scriptconfig as scfg


# FIXME: Looks like this CLI is not functional, which might be fine considering
# this is meant to be used as a library.

class SmartflowEgressConfig(scfg.DataConfig):
    """
    Egress KWCOCO data to T&E baseline framework structure
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    assets = scfg.Value(None, type=str, position=2, required=True, help='Assets to upload', nargs='+')
    outdir = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            Output directory for ingressed assets an output STAC Catalog
            '''))
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use. UNUSED. Hook up to fsspec if needed.
            '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='UNUSED. DEPRECATED')
    show_progress = scfg.Value(False, isflag=True, short_alias=['s'], help='UNUSED. DEPRECATED')


def main():
    smartflow_egress_with_arg_processing(**SmartflowEgressConfig.cli())


def _build_stac_item(region_path,
                     assetnames_and_s3_paths):
    with open(region_path) as f:
        data = json.load(f)

    from geowatch.geoannots.geomodels import RegionModel
    region = RegionModel(**data)

    # These are fast checks that include the assertion that there is only one
    # header (i.e. type=region) feature.
    region._validate_quick_checks()

    import shapely
    region_geometry: shapely.geometry.polygon.Polygon = region.geometry
    region_bbox = list(region_geometry.bounds)

    return {'type': 'Feature',
            'stac_version': '1.0.0',
            'stac_extensions': [],
            'id': uuid.uuid4().hex,
            'geometry': shapely.geometry.mapping(region_geometry),
            'bbox': region_bbox,
            'properties': {},
            'assets': assetnames_and_s3_paths}


def smartflow_egress_with_arg_processing(
        assetnames_and_paths,
        region_path,
        output_path,
        outbucket,
        aws_profile=None,
        dryrun=False,
        newline=False,
        show_progress=False):
    assetnames_and_local_paths = {}
    for assetname_and_path in assetnames_and_paths:
        asset, local_path = assetname_and_path.split(':')

        assetnames_and_local_paths[asset] = local_path

    return smartflow_egress(assetnames_and_local_paths,
                            region_path,
                            output_path,
                            outbucket,
                            aws_profile=aws_profile,
                            dryrun=dryrun,
                            newline=newline,
                            show_progress=show_progress)


def smartflow_egress(assetnames_and_local_paths,
                     region_path,
                     output_path,
                     outbucket,
                     aws_profile=None,
                     dryrun=False,
                     newline=False,
                     show_progress=False):
    """
    Uploads specified assets to S3 with a STAC manifest.

    Args:
        assetnames_and_local_paths (Dict):
            Mapping from an asset name to the local path to upload. The asset
            name will be indexable in the uploaded STAC item. Any local path
            specified multiple times will only be uploaded once, but multiple
            STAC assets will be associated with it.

        region_path (str | PathLike):
            local path to the region file associated with a processing node

        output_path (str):
            The path in the s3 bucket that the stac item will be uploaded to.

        outbucket (str):
            The s3 bucket that assets will be uploaded to.

        aws_profile (str | None): aws cp argument

        newline (bool): controls formatting of output stac item

    Returns:
        Dict: Of asset names and output paths

    CommandLine:
        xdoctest -m geowatch.cli.smartflow_egress smartflow_egress

    Example:
        >>> from geowatch.cli.smartflow_egress import *  # NOQA
        >>> from geowatch.geoannots.geomodels import RegionModel
        >>> from os.path import join
        >>> dpath = ub.Path.appdir('geowatch/tests/smartflow_egress').ensuredir()
        >>> local_dpath = (dpath / 'local').ensuredir()
        >>> remote_root = (dpath / 'fake_s3_loc').ensuredir()
        >>> #outbucket = 's3://fake/bucket'
        >>> outbucket = remote_root
        >>> output_path = join(outbucket, 'items.jsonl')
        >>> region = RegionModel.random()
        >>> region_path = dpath / 'demo_region.geojson'
        >>> region_path.write_text(region.dumps())
        >>> assetnames_and_local_paths = {
        >>>     'asset_file1': dpath / 'my_path1.txt',
        >>>     'asset_file2': dpath / 'my_path2.txt',
        >>>     'asset_file_reference': dpath / 'my_path1.txt',
        >>>     'asset_dir1': dpath / 'my_dir1',
        >>> }
        >>> # Generate local data we will pretend to egress
        >>> assetnames_and_local_paths['asset_file1'].write_text('foobar1')
        >>> assetnames_and_local_paths['asset_file2'].write_text('foobar2')
        >>> assetnames_and_local_paths['asset_dir1'].ensuredir()
        >>> (assetnames_and_local_paths['asset_dir1'] / 'data1').write_text('data1')
        >>> (assetnames_and_local_paths['asset_dir1'] / 'data1').write_text('data2')
        >>> te_output = smartflow_egress(
        >>>     assetnames_and_local_paths,
        >>>     region_path,
        >>>     output_path,
        >>>     outbucket,
        >>>     newline=False,
        >>> )

    Ignore:
        >>> # Requires a real S3 bucket
        >>> from geowatch.cli.smartflow_egress import *  # NOQA
        >>> from geowatch.geoannots.geomodels import RegionModel
        >>> from geowatch.utils import util_fsspec
        >>> from os.path import join
        >>> dpath = ub.Path.appdir('geowatch/tests/smartflow_egress').ensuredir()
        >>> local_dpath = (dpath / 'local').ensuredir()
        >>> remote_root = (dpath / 'fake_s3_loc').ensuredir()
        >>> # A REAL AWS PATH WE HAVE ACCESS TO
        >>> outbucket = util_fsspec.S3Path.coerce('s3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/tests/test-egress')
        >>> if 0:
        >>>     outbucket.delete()
        >>> output_path = join(outbucket, 'items.jsonl')
        >>> region = RegionModel.random()
        >>> region_path = dpath / 'demo_region.geojson'
        >>> region_path.write_text(region.dumps())
        >>> assetnames_and_local_paths = {
        >>>     'asset_file1': dpath / 'my_path1.txt',
        >>>     'asset_file2': dpath / 'my_path2.txt',
        >>>     'asset_file_reference': dpath / 'my_path1.txt',
        >>>     'asset_dir1': dpath / 'my_dir1',
        >>> }
        >>> # Generate local data we will pretend to egress
        >>> assetnames_and_local_paths['asset_file1'].write_text('foobar1')
        >>> assetnames_and_local_paths['asset_file2'].write_text('foobar2')
        >>> assetnames_and_local_paths['asset_dir1'].ensuredir()
        >>> (assetnames_and_local_paths['asset_dir1'] / 'data1').write_text('data1')
        >>> (assetnames_and_local_paths['asset_dir1'] / 'data1').write_text('data2')
        >>> te_output = smartflow_egress(
        >>>     assetnames_and_local_paths,
        >>>     region_path,
        >>>     output_path,
        >>>     outbucket,
        >>>     newline=False,
        >>> )
        >>> outbucket.ls()
        >>> (outbucket / 'my_dir1').ls()
        >>> # Test subsequent ingress
        >>> from geowatch.cli.smartflow_ingress import smartflow_ingress
        >>> in_dpath = ub.Path.appdir('geowatch/tests/smartflow_ingress2').delete().ensuredir()
        >>> input_path = output_path
        >>> assets = ['asset_file1', 'asset_dir1']
        >>> kwcoco_stac_item_assets = smartflow_ingress(
        >>>     input_path,
        >>>     assets,
        >>>     in_dpath,
        >>> )
    """
    # TODO: handle aws_profile.
    from geowatch.utils.util_fsspec import FSPath
    print('--- BEGIN EGRESS ---')
    print(f'outbucket   = {outbucket}')
    print(f'output_path = {output_path}')

    assert aws_profile is None, 'unhandled'
    outbucket = FSPath.coerce(outbucket)

    PRE_DELETE_HACK = 0
    if PRE_DELETE_HACK:
        # HACK: delete everything in the outbucket to prevent conflicting
        # results and ensure the next step always gets exactly this output
        # and nothing more.
        if outbucket.exists():
            print(f'DELETE EXISTING: outbucket={outbucket}')
            outbucket.delete()
        else:
            print('Outbucket doesnt exist yet')

    # Make a temporary output path for a transactional upload.
    tmp_prefix = 'tmp-' + ub.timestamp() + '-' + ub.hash_data(uuid.uuid4())[0:8] + '-'
    tmp_parent = (outbucket.parent / 'tmp').ensuredir()
    tmp_outbucket = tmp_parent / (tmp_prefix + outbucket.name)
    tmp_outbucket.ensuredir()

    # TODO: Can use fsspec to grab multiple files in parallel
    assetnames_and_s3_paths = {}

    items = list(assetnames_and_local_paths.items())
    seen = set()  # Prevent duplicate uploads
    for asset, local_path in ub.ProgIter(items, desc='Egress data', verbose=3):
        local_path = FSPath.coerce(local_path)
        if local_path.startswith('s3'):
            # Assets with paths already on S3 simply pass a reference through
            final_asset_s3_outpath = local_path
            tmp_asset_s3_outpath = None
        else:
            # Otherwise do a copy. Mark the temporary transaction location
            # and the real final location.
            final_asset_s3_outpath = outbucket / local_path.name
            tmp_asset_s3_outpath = tmp_outbucket / local_path.name
            if local_path not in seen:
                fallback_copy(local_path, tmp_asset_s3_outpath)
                # local_path.copy(asset_s3_outpath)
                seen.add(local_path)

        assetnames_and_s3_paths[asset] = {'href': str(final_asset_s3_outpath)}

    output_stac_item = _build_stac_item(region_path,
                                        assetnames_and_s3_paths)

    # Only have the one KWCOCO dataset STAC item in this case
    output_stac_items = [output_stac_item]
    if newline:
        te_output = '\n'.join((json.dumps(item) for item in output_stac_items))
    else:
        te_output = {'raw_images': [],
                     'stac': {
                         'type': 'FeatureCollection',
                         'features': output_stac_items}}

    # Finish transaction
    tmp_outbucket.move(outbucket)

    # Write the final file after the move because it often will write into the
    # final directory.
    with tempfile.NamedTemporaryFile() as temporary_file:
        with open(temporary_file.name, 'w') as f:
            if newline:
                print(te_output, end='', file=f)
            else:
                print(json.dumps(te_output, indent=2), file=f)

        _temp_path = FSPath.coerce(temporary_file.name)
        _output_path = FSPath.coerce(output_path)
        fallback_copy(_temp_path, _output_path)

    print('EGRESSED: {}'.format(ub.urepr(output_stac_items, nl=-1)))
    print('--- FINISH EGRESS ---')
    return {k: v['href'] for k, v in assetnames_and_s3_paths.items()}


def fallback_copy(local_path, asset_s3_outpath):
    """
    Copying with fsspec alone seems to be causing issues.
    This provides a fallback to a raw S3 command, as well as other verbosity.
    """
    from geowatch.utils import util_fsspec
    from geowatch.utils import util_framework
    assert isinstance(local_path, util_fsspec.LocalPath)

    DO_FALLBACK = 1

    # callback seems to break, not sure why, fixme?
    # from fsspec.callbacks import TqdmCallback
    # callback = TqdmCallback(tqdm_kwargs={"desc": "Copying"})
    if local_path.is_dir() and isinstance(asset_s3_outpath, util_fsspec.S3Path):

        HACK_ENSURE_NON_EMPTY_DIRS = True
        if HACK_ENSURE_NON_EMPTY_DIRS:
            # If the directory is empty, write a dummy file to force it to
            # upload to s3.
            import os
            if not any(os.scandir(local_path)):
                (local_path  / '__dir__').write_text('Hack to ensure directory is non-empty')

        if DO_FALLBACK:
            try:
                local_path.copy(asset_s3_outpath, verbose=3)
            except Exception:
                print('fsspec copy failed, fallback to aws CLI')
                # In the case where we are moving a directory from the local to
                # s3 we *should* just be able to use copy, but because that
                # seems to be breaking, we are falling back on an explicit aws
                # cli command
                profile = asset_s3_outpath.fs.storage_options.get('profile', None)
                aws_kwargs = {}
                if profile is not None:
                    aws_kwargs['profile'] = profile
                aws_cmd = util_framework.AWS_S3_Command(
                    'sync', local_path, asset_s3_outpath, **aws_kwargs)
                aws_cmd.run()
        else:
            local_path.copy(asset_s3_outpath, verbose=3)
            #callback=callback)
    else:
        # In every other case, regular copy is probably fine
        local_path.copy(asset_s3_outpath, verbose=3)
        #callback=callback)


if __name__ == "__main__":
    main()



"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/time_kernel_grammar.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import MultiTimeKernelTransformer  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import _global_multi_time_kernel_parser  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import parse_multi_time_kernel  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import MULTI_TIME_KERNEL_GRAMMAR  # NOQA



#!/usr/bin/env python3
"""
See Old Version:
    ../../../scripts/run_bas_fusion_eval3_for_baseline.py

SeeAlso:
    ~/code/watch-smartflow-dags/KIT_TA2_PREEVAL10_PYENV.py
"""
import json
import os
import scriptconfig as scfg
import shutil
import ubelt as ub
from datetime import datetime, timezone
from dateutil import parser


class BasFusionConfig(scfg.DataConfig):
    """
    Run TA-2 BAS fusion as baseline framework component

    python ~/code/watch/geowatch/cli/smartflow/run_bas_fusion.py
    """

    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help='S3 path for output JSON')

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands.
            '''))

    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='DEPRECATED. DO NOT USE.')

    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))

    models_outbucket = scfg.Value(None, type=str, required=False, help=ub.paragraph(
            '''
            S3 Output directory for output region and site models (if
            not specified, defaults to using the `outbucket`
            parameter)
            '''))

    ta2_s3_collation_bucket = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            S3 Location for collated TA-2 output (bucket name should
            include up to eval name)
            '''))

    previous_bas_outbucket = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            S3 Output directory for previous interval BAS fusion output
            '''))

    bas_pxl_config = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            Raw json/yaml or a path to a json/yaml file that specifies the
            config for fusion.predict.
            '''))

    bas_poly_config = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            Raw json/yaml or a path to a json/yaml file that specifies the
            config for bas tracking.
            '''))

    previous_interval_output = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            Output path for previous interval BAS DatasetGen step
            '''))

    num_years_historical = scfg.Value(None, type=int, help=ub.paragraph(
            '''
            Number of years worth of historical data to consider; does
            nothing if 'previous_interval_output' is not specified
            (i.e. for incremental mode)
            '''))

    time_dense = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Use time_dense imagery. Defaults to False and uses time averaged data.
            '''))

    egress_intermediate_outputs = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        If true egress intermediate heatmaps, otherwise only egress the geojson
        '''))


__debug_notes__ = r"""
config = {
    'input_path'             : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v130/batch/kit/KR_R002/2021-08-31/split/mono/products/cold/items.jsonl',
    'input_region_path'      : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v130/batch/kit/KR_R002/2021-08-31/input/mono/region_models/KR_R002.geojson',
    'output_path'            : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v130/batch/kit/KR_R002/2021-08-31/split/mono/products/bas-fusion/items.jsonl',
    'aws_profile'            : None,
    'dryrun'                 : False,
    'outbucket'              : 's3://smartflow-023300502152-us-west-2/smartflow/env/kw-v3-0-0/work/preeval17_batch_v130/batch/kit/KR_R002/2021-08-31/split/mono/products/bas-fusion',
    'ta2_s3_collation_bucket': None,
    'previous_bas_outbucket' : None,
    'bas_pxl_config'         : 'chip_dims: auto\nchip_overlap: 0.3\nfixed_resolution: 10GSD\nnum_workers: 24\npackage_fpath: /root/data/smart_expt_dvc/models/fusion/uconn/D7-V2-COLD-candidate/epoch=203-step=4488.pt\ntime_sampling: soft4\ntime_span: auto\ntta_fliprot: 3\ntta_time: 3',
    'bas_poly_config'        : 'agg_fn: probs\ninner_agg_fn: mean\ninner_window_size: 1y\nmax_area_square_meters: 8000000\nmin_area_square_meters: 7200\nmoving_window_size: null\nnorm_ord: inf\npoly_merge_method: v2\npolygon_simplify_tolerance: 1\nresolution: 10GSD\nthresh: 0.375\ntime_thresh: 0.8',
}
config = BasFusionConfig(**config)

"""


def main():
    config = BasFusionConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_bas_fusion_for_baseline(config)


def filter_kwcoco_images_by_datetime(input_kwcoco_fpath,
                                     output_kwcoco_fpath,
                                     date,
                                     mode='remove-after'):
    assert mode in {'remove-after', 'remove-before'}

    import kwcoco
    from operator import lt, ge

    if mode == 'remove-before':
        op = lt
    elif mode == 'remove-after':
        op = ge
    else:
        raise RuntimeError("Expecting kwarg `mode` to be one "
                           "of: {'remove-before', 'remove-after'}")

    input_dset = kwcoco.CocoDataset(input_kwcoco_fpath)

    image_ids_to_remove = []
    for o in input_dset.images().objs:
        odate = parser.parse(o['date_captured'])

        if op(odate, date):
            image_ids_to_remove.append(o['id'])

    input_dset.remove_images(image_ids_to_remove)

    input_dset.dump(output_kwcoco_fpath)


def run_bas_fusion_for_baseline(config):
    from geowatch.cli.smartflow_ingress import smartflow_ingress
    from geowatch.cli.smartflow_egress import smartflow_egress
    from geowatch.cli.concat_kwcoco_videos import concat_kwcoco_datasets
    from geowatch.utils.util_framework import download_region
    from geowatch.tasks.fusion.predict import predict
    from kwutil.util_yaml import Yaml
    from geowatch.utils import util_framework
    from geowatch.utils import util_fsspec

    ####
    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()

    input_path = config.input_path
    input_region_path = config.input_region_path
    outbucket = config.outbucket
    aws_profile = config.aws_profile
    dryrun = config.dryrun
    previous_bas_outbucket = config.previous_bas_outbucket
    ta2_s3_collation_bucket = config.ta2_s3_collation_bucket

    if aws_profile is not None:
        # This should be sufficient, but it is not tested.
        util_fsspec.S3Path._new_fs(profile=aws_profile)

    assert not dryrun, 'unsupported'

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = ub.Path('/tmp/ingress')
    assets = [
        'enriched_bas_kwcoco_file',
        'enriched_bas_kwcoco_teamfeats',
        'enriched_bas_kwcoco_rawbands',
        {"key": 'hacked_cold_assets', "allow_missing": True},
        {"key": 'landcover_assets', "allow_missing": True},
    ]
    if config.time_dense:
        assets += [
            'timedense_bas_kwcoco_file',
            'timedense_bas_kwcoco_rawbands',
        ]

    ingressed_assets = smartflow_ingress(
        input_path=input_path,
        assets=assets,
        outdir=ingress_dir,
        aws_profile=aws_profile,
        dryrun=dryrun
    )

    # # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'

    download_region(
        input_region_path=input_region_path,
        output_region_path=local_region_path,
        aws_profile=aws_profile,
        strip_nonregions=True,
        ensure_comments=True,
    )

    from geowatch.geoannots.geomodels import RegionModel
    region = RegionModel.coerce(local_region_path)

    # Returned as datetime
    current_interval_end_date = region.end_date

    if current_interval_end_date.month == 1 and current_interval_end_date.day == 1:
        # If current interval ends at the start of a year,
        # consider the "current" year to be the previous one
        current_interval_year = current_interval_end_date.year - 1
    else:
        current_interval_year = current_interval_end_date.year

    # Granularity of filtering only by year currently since we're
    # using data time-averaged over a year
    if config.num_years_historical is None:
        min_date = datetime(
            current_interval_year,
            1, 1, tzinfo=timezone.utc)
    else:
        min_date = datetime(
            current_interval_year - (config.num_years_historical - 1),
            1, 1, tzinfo=timezone.utc)

    # Determine the region_id in the region file.
    region_id = region.region_id

    # 3. Run fusion
    print("* Running BAS fusion *")
    bas_fusion_kwcoco_path = ingress_dir / 'bas_fusion_kwcoco.json'

    node_state.print_current_state(ingress_dir)

    # TODO: remove these defaults or replace them with whatever is the
    # default in predict. The params should be fully given in the DAG, not
    # here.
    default_predict_config = ub.udict({
          "chip_overlap": 0.3,
          "chip_dims": "auto",
          "time_span": "auto",
          "time_sampling": "auto",
          "drop_unused_frames": True,
          "batch_size": 1,
          "num_workers": 2,
          'package_fpath': None,
    })
    bas_pxl_config = default_predict_config | Yaml.coerce(config.bas_pxl_config or {})

    if bas_pxl_config.get('package_fpath', None) is None:
        raise ValueError('Requires package_fpath')

    print('bas_pxl_config = {}'.format(ub.urepr(bas_pxl_config, nl=1)))

    ingress_kwcoco_path = ingressed_assets['enriched_bas_kwcoco_file']

    incremental_assets_for_egress = {}
    if config.previous_interval_output is None:
        filtered_ingress_kwcoco_path = ingress_kwcoco_path
    else:
        filtered_ingress_kwcoco_path =\
            ingress_dir / 'filtered_enriched_kwcoco.json'

        filter_kwcoco_images_by_datetime(
            ingress_kwcoco_path,
            filtered_ingress_kwcoco_path,
            min_date, mode='remove-before')

        incremental_assets_for_egress['filtered_enriched_kwcoco'] =\
            filtered_ingress_kwcoco_path

    if 0:
        import kwcoco
        src_dset = kwcoco.CocoDataset(ingress_kwcoco_path)
        src_dset.validate()

    predict(devices='0,',
            write_preds=False,
            write_probs=True,
            with_change=False,
            with_saliency=True,
            with_class=False,
            test_dataset=filtered_ingress_kwcoco_path,
            pred_dataset=bas_fusion_kwcoco_path,
            **bas_pxl_config)

    node_state.print_current_state(ingress_dir)

    # 3.1. If a previous interval was run; concatenate BAS fusion
    # output KWCOCO files for tracking
    if previous_bas_outbucket is not None:
        combined_bas_fusion_kwcoco_path = ingress_dir / 'combined_bas_fusion_kwcoco.json'

        previous_ingress_dir = ub.Path('/tmp/ingress_previous')

        previous_bas_outbucket = util_fsspec.FSPath.coerce(previous_bas_outbucket)
        previous_ingress_dir = util_fsspec.FSPath.coerce(previous_ingress_dir)
        previous_bas_outbucket.copy(previous_ingress_dir, recursive=True)

        previous_bas_fusion_kwcoco_path = previous_ingress_dir / 'combined_bas_fusion_kwcoco.json'

        # On first interval nothing will be copied down so need to
        # check that we have the input explicitly
        if os.path.isfile(previous_bas_fusion_kwcoco_path):
            concat_kwcoco_datasets(
                (previous_bas_fusion_kwcoco_path, bas_fusion_kwcoco_path),
                combined_bas_fusion_kwcoco_path)
            # Copy saliency assets from previous bas fusion
            shutil.copy_tree(
                previous_ingress_dir / '_assets/pred_saliency',
                ingress_dir / '_assets/pred_saliency'
            )

            # Copy original assets from previous bas rusion
            shutil.copy_tree(
                previous_ingress_dir / region_id,
                ingress_dir / region_id
            )
        else:
            # Copy current bas_fusion_kwcoco_path to combined path as
            # this is the first interval
            shutil.copy(bas_fusion_kwcoco_path,
                        combined_bas_fusion_kwcoco_path)
    else:
        combined_bas_fusion_kwcoco_path = bas_fusion_kwcoco_path

    previous_ingressed_assets = None
    if config.previous_interval_output is not None:
        print('* Combining previous interval time combined kwcoco with '
              'current *')
        previous_ingress_dir = ub.Path('/tmp/ingress_previous')
        try:
            previous_ingressed_assets = smartflow_ingress(
                config.previous_interval_output,
                ['bas_pred_saliency_assets',
                 'combined_bas_fusion_kwcoco_path'],
                previous_ingress_dir,
                config.aws_profile,
                config.dryrun)
        except FileNotFoundError:
            print("** Warning: Couldn't ingress previous interval output; "
                  "assuming this is the first interval **")
            combined_kwcoco_path = combined_bas_fusion_kwcoco_path
        else:
            combined_kwcoco_path = ingress_dir / 'combined_bas_fusion_kwcoco.json'

            filtered_previous_bas_fusion_kwcoco_path =\
                previous_ingress_dir / 'filtered_bas_fusion_kwcoco.json'

            filter_kwcoco_images_by_datetime(
                previous_ingressed_assets['combined_bas_fusion_kwcoco_path'],
                filtered_previous_bas_fusion_kwcoco_path,
                min_date, mode='remove-after')

            incremental_assets_for_egress['filtered_bas_fusion_kwcoco'] =\
                filtered_previous_bas_fusion_kwcoco_path

            from geowatch.cli.concat_kwcoco_videos import concat_kwcoco_datasets
            concat_kwcoco_datasets(
                (filtered_previous_bas_fusion_kwcoco_path,
                 combined_bas_fusion_kwcoco_path),
                combined_kwcoco_path)

            # Copy saliency assets from previous bas fusion
            shutil.copytree(previous_ingress_dir / 'pred_saliency',
                            ingress_dir / '_assets' / 'pred_saliency',
                            dirs_exist_ok=True)
    else:
        combined_kwcoco_path = combined_bas_fusion_kwcoco_path

    # 4. Compute tracks (BAS)
    print("* Computing tracks (BAS) *")
    bas_region_models_outdir = (ingress_dir / 'region_models').ensuredir()
    bas_site_models_outdir = (ingress_dir / 'site_models_bas').ensuredir()
    region_models_manifest_outdir = (ingress_dir / 'tracking_manifests_bas').ensuredir()
    region_models_manifest_outpath = region_models_manifest_outdir / 'region_models_manifest.json'
    site_models_manifest_outpath = region_models_manifest_outdir / 'site_models_manifest.json'

    # Copy input region model into region_models outdir to be updated
    # (rather than generated from tracking, which may not have the
    # same bounds as the original)
    shutil.copy(local_region_path, bas_region_models_outdir / f'{region_id}.geojson')

    # TODO: remove these defaults or replace them with whatever is the
    # default in tracker. The params should be fully given in the DAG,
    # not here.
    default_bas_tracking_config = ub.udict({
        "thresh": 0.1,
        "moving_window_size": None,
        "polygon_simplify_tolerance": 1,
        "max_area_behavior": 'ignore'
    })
    bas_tracking_config = (default_bas_tracking_config
                           | Yaml.coerce(config.bas_poly_config or {}))

    min_area_square_meters = bas_tracking_config.get('min_area_square_meters', None)
    time_pad_after = bas_tracking_config.pop('time_pad_after', None)
    time_pad_before = bas_tracking_config.pop('time_pad_before', None)
    # TODO: use smart_pipeline.BAS_PolygonPrediction
    tracked_bas_kwcoco_path = '_tracked'.join(
        os.path.splitext(combined_kwcoco_path))
    ub.cmd([
        'python', '-m', 'geowatch.cli.run_tracker',
        combined_kwcoco_path,
        '--out_sites_dir', bas_site_models_outdir,
        '--out_sites_fpath', site_models_manifest_outpath,
        '--out_site_summaries_dir', bas_region_models_outdir,
        '--out_site_summaries_fpath', region_models_manifest_outpath,
        '--out_kwcoco', tracked_bas_kwcoco_path,
        '--default_track_fn', 'saliency_heatmaps',
        '--append_mode', 'True',
        '--time_pad_after', str(time_pad_after),
        '--time_pad_before', str(time_pad_before),
        # TODO:
        # use boundary_region here?
        '--track_kwargs', json.dumps(bas_tracking_config)],
        check=True, verbose=3)

    # Remove after boundary_region here
    cropped_region_models_outdir = (ingress_dir / 'cropped_region_models_bas').ensuredir()
    cropped_site_models_outdir = (ingress_dir / 'cropped_site_models_bas').ensuredir()

    node_state.print_current_state(ingress_dir)

    crop_cmd = [
        'python', '-m', 'geowatch.cli.crop_sites_to_regions',
        '--site_models', bas_site_models_outdir / '*.geojson',
        '--region_models', bas_region_models_outdir / '*.geojson',
        '--new_site_dpath', cropped_site_models_outdir,
        '--new_region_dpath', cropped_region_models_outdir,
        '--min_area_square_meters', str(min_area_square_meters)
    ]
    ub.cmd(crop_cmd, check=True, verbose=3)

    # Validate and fix all outputs
    try:
        util_framework.fixup_and_validate_site_and_region_models(
            region_dpath=cropped_region_models_outdir,
            site_dpath=cropped_site_models_outdir,
        )
    except Exception as ex:  # noqa
        print(f'Encountered Exception, ex={ex}, uploading debug informaiton and then exiting')
        debug_s3_outdir = os.path.join(outbucket, '_debug/')
        cropped_region_models_outdir = util_fsspec.FSPath.coerce(cropped_region_models_outdir)
        debug_s3_outdir = util_fsspec.FSPath.coerce(debug_s3_outdir)
        cropped_region_models_outdir.copy(debug_s3_outdir, recursive=True)
        cropped_site_models_outdir.copy(debug_s3_outdir, recursive=True)
        raise

    if __debug__:
        from geowatch.geoannots import geomodels
        all_regions_models = list(geomodels.RegionModel.coerce_multiple(cropped_region_models_outdir))
        all_site_models = list(geomodels.SiteModel.coerce_multiple(cropped_site_models_outdir))
        assert len(all_regions_models) == 1, 'should only be 1 output region model'
        out_region_model = all_regions_models[0]
        num_site_summaries = len(list(out_region_model.site_summaries()))
        num_sites = len(all_site_models)
        assert num_sites == num_site_summaries, 'number of site summaries should be the same as the number of sites'

    # 6. (Optional) collate TA-2 output
    if ta2_s3_collation_bucket is not None:
        # only used if we are going to submit our BAS predictions as the final
        # ones?
        print("* Collating TA-2 output")
        util_framework.ta2_collate_output(None,
                                          cropped_region_models_outdir,
                                          cropped_site_models_outdir,
                                          ta2_s3_collation_bucket)

    EGRESS_INTERMEDIATE_OUTPUTS = config.egress_intermediate_outputs
    if EGRESS_INTERMEDIATE_OUTPUTS:
        # Reroot kwcoco files to make downloaded results easier to work with
        ub.cmd(['kwcoco', 'reroot', f'--src={bas_fusion_kwcoco_path}', '--inplace=1', '--absolute=0'])
        ub.cmd(['kwcoco', 'reroot', f'--src={tracked_bas_kwcoco_path}', '--inplace=1', '--absolute=0'])
        # Add BAS saliency outputs to egressed attributes for debugging
        ingressed_assets['bas_pred_saliency_assets'] = ingress_dir / '_assets/pred_saliency'
        ingressed_assets['bas_fusion_kwcoco_path'] = bas_fusion_kwcoco_path
        ingressed_assets['combined_bas_fusion_kwcoco_path'] = combined_kwcoco_path
        ingressed_assets['bas_original_site_models_outdir'] = bas_site_models_outdir
        ingressed_assets['bas_original_region_models_outdir'] = bas_region_models_outdir
        ingressed_assets['tracked_bas_kwcoco_path'] = tracked_bas_kwcoco_path

    node_state.print_current_state(ingress_dir)

    models_to_egress = {'cropped_region_models_bas': cropped_region_models_outdir,
                        'cropped_site_models_bas': cropped_site_models_outdir}
    if config.models_outbucket is not None:
        # Egress region models to a dedicated bucket
        models_egressed = smartflow_egress(
            models_to_egress,
            local_region_path,
            os.path.join(config.models_outbucket, 'models_only_items.jsonl'),
            config.models_outbucket,
            aws_profile=None,
            dryrun=False,
            newline=False)

        ingressed_assets = {**ingressed_assets, **models_egressed}
    else:
        ingressed_assets = {**ingressed_assets, **models_to_egress}

    # 6. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    smartflow_egress(ingressed_assets,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=None,
                     dryrun=False,
                     newline=False)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class FixupConfig(scfg.DataConfig):
    """
    Given the final results of the system, run our validation scripts and any
    final cleanups that need to happen before submitting to T&E validation.
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
        '''
        Path to the STAC items this step can use as inputs.
        This is usually an S3 Path.
        '''), alias=['input_stac_path'])

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
        '''
        Path to input T&E Baseline Framework Region definition JSON
        '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help=ub.paragraph(
        '''
        Path to the STAC items that register the outputs of this stage.
        This is usually an S3 Path.
        '''), alias=['output_stac_path'])

    ta2_s3_collation_bucket = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        S3 Location for collated TA-2 output (bucket name should
        include up to eval name)
        '''))

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        AWS Profile to use for AWS S3 CLI commands
        '''))

    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='DEPRECATED.')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
        '''
        S3 Output directory for STAC item / asset egress
        '''))

    input_region_models_asset_name = scfg.Value('cropped_region_models_sc', type=str, required=False, help=ub.paragraph(
        '''
        Which region model assets to ingress and fix up
        '''), alias=['region_models_asset_name'])

    input_site_models_asset_name = scfg.Value('cropped_site_models_sc', type=str, required=False, help=ub.paragraph(
        '''
        Which site model assets to ingress and fix up
        '''), alias=['site_models_asset_name'])

    performer_suffix = scfg.Value('KIT', type=str, required=False, help=ub.paragraph(
            '''
            Performer suffix for output files
            '''))


def main(cmdline=1, **kwargs):
    config = FixupConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    print('config = {}'.format(ub.urepr(config, nl=1, align=':')))

    from geowatch.utils.util_framework import NodeStateDebugger
    node_state = NodeStateDebugger()
    node_state.print_environment()
    node_state.print_local_invocation(config)

    from geowatch.cli.smartflow_ingress import smartflow_ingress
    from geowatch.cli.smartflow_egress import smartflow_egress
    from geowatch.utils.util_framework import download_region
    from geowatch.utils import util_framework

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = ub.Path('/tmp/ingress')

    if not config.input_site_models_asset_name:
        raise ValueError(ub.paragraph(
            '''
            Must specify which asset name in the STAC catalog points to the
            site models
            '''))

    if not config.input_region_models_asset_name:
        raise ValueError(ub.paragraph(
            '''
            Must specify which asset name in the STAC catalog points to the
            site models
            '''))

    input_path = config.input_path
    assets = [
        {'key': config.input_region_models_asset_name},
        {'key': config.input_site_models_asset_name, 'missing_action': 'mkdir'},
    ]
    outdir = ingress_dir
    aws_profile = config.aws_profile
    dryrun = config.dryrun
    # show_progress = False
    # dont_error_on_missing_asset = False

    ingressed_assets = smartflow_ingress(
        input_path,
        assets,
        outdir,
        aws_profile,
        dryrun)

    # # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'

    download_region(
        input_region_path=config.input_region_path,
        output_region_path=local_region_path,
        aws_profile=config.aws_profile,
        strip_nonregions=True,
    )

    node_state.print_current_state(ingress_dir)

    dummy_kwcoco_path = ingress_dir / 'dummy.kwcoco.json'
    dummy_kwcoco_path.touch()

    input_region_dpath = ub.Path(ingressed_assets[config.input_region_models_asset_name])
    input_site_dpath = ub.Path(ingressed_assets[config.input_site_models_asset_name])

    output_region_dpath = ingress_dir / 'cropped_region_models_fixed'
    output_site_dpath = ingress_dir / 'cropped_site_models_fixed'

    # Copy the input to the output because we will modify them inplace.
    input_site_dpath.copy(output_site_dpath)
    input_region_dpath.copy(output_region_dpath)

    # Validate and fix all outputs
    print('Fixup and validate outputs')
    util_framework.fixup_and_validate_site_and_region_models(
        region_dpath=output_region_dpath,
        site_dpath=output_site_dpath,
    )

    node_state.print_current_state(ingress_dir)

    # 5. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    ingressed_assets['cropped_region_models_fixed'] = output_region_dpath
    ingressed_assets['cropped_site_models_fixed'] = output_site_dpath
    smartflow_egress(ingressed_assets,
                     local_region_path,
                     config.output_path,
                     config.outbucket,
                     aws_profile=config.aws_profile,
                     dryrun=False,
                     newline=False)

    # 6. (Optional) collate TA-2 output
    if config.ta2_s3_collation_bucket is not None:
        print("* Collating TA-2 output")
        util_framework.ta2_collate_output(None,
                                          output_region_dpath,
                                          output_site_dpath,
                                          config.ta2_s3_collation_bucket,
                                          performer_suffix=config.performer_suffix)


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/geowatch/cli/smartflow/run_fixups.py
    """
    main()
